from functools import lru_cache
from typing import List, Union

import pandas as pd
import torch
from chromadb import Settings
from langchain.callbacks.streaming_stdout import StreamingStdOutCallbackHandler
from langchain.docstore.document import Document
from langchain.document_loaders import DataFrameLoader
from langchain.embeddings import HuggingFaceEmbeddings
from langchain.vectorstores import Chroma
from pydantic import BaseModel

DEFAULT_EMBEDDINGS_MODEL = "sentence-transformers/all-mpnet-base-v2"
USER_DEFINED_MODEL_PARAMS = (
    "model_name",
    "max_tokens",
    "temperature",
    "top_p",
    "stop",
    "best_of",
    "verbose",
    "writer_org_id",
    "writer_api_key",
    "base_url",
)


class ModelParameters(BaseModel):
    """Model parameters for the Writer LLM API interface"""

    writer_api_key: str = None
    writer_org_id: str = None
    base_url: str = None
    model_id: str = "palmyra-x"
    callbacks: List[StreamingStdOutCallbackHandler] = [StreamingStdOutCallbackHandler()]
    max_tokens: int = 1024
    temperature: float = 0.0
    top_p: float = 1
    stop: List[str] = []
    best_of: int = 5
    verbose: bool = False

    class Config:
        arbitrary_types_allowed = True


class DfLoader(DataFrameLoader):

    """
    override the load method of langchain.document_loaders.DataFrameLoaders to ignore rows with 'None' values
    """

    def __init__(self, data_frame: pd.DataFrame, page_content_column: str):
        super().__init__(data_frame=data_frame, page_content_column=page_content_column)
        self._data_frame = data_frame
        self._page_content_column = page_content_column

    def load(self) -> List[Document]:
        """Loads the dataframe as a list of documents"""
        documents = []
        for n_row, frame in self._data_frame[self._page_content_column].iteritems():
            if pd.notnull(frame):
                # ignore rows with None values
                column_name = self._page_content_column

                document_contents = frame

                documents.append(
                    Document(
                        page_content=document_contents,
                        metadata={
                            "source": "dataframe",
                            "row": n_row,
                            "column": column_name,
                        },
                    )
                )
        return documents


def df_to_documents(
    df: pd.DataFrame, page_content_columns: Union[List[str], str]
) -> List[Document]:
    """Converts a given dataframe to a list of documents"""
    documents = []

    if isinstance(page_content_columns, str):
        page_content_columns = [page_content_columns]

    for _, page_content_column in enumerate(page_content_columns):
        if page_content_column not in df.columns.tolist():
            raise ValueError(
                f"page_content_column {page_content_column} not in dataframe columns"
            )

        loader = DfLoader(data_frame=df, page_content_column=page_content_column)
        documents.extend(loader.load())

    return documents


@lru_cache()
def load_embeddings_model(embeddings_model_name):
    try:
        model_kwargs = {"device": "gpu" if torch.cuda.is_available() else "cpu"}
        embedding_model = HuggingFaceEmbeddings(
            model_name=embeddings_model_name, model_kwargs=model_kwargs
        )
    except ValueError:
        raise ValueError(
            f"The {embeddings_model_name}  is not supported, please select a valid option from Hugging Face Hub!"
        )
    return embedding_model


def load_chroma(
    embeddings_model_name, persist_directory, collection_name, chroma_settings
):
    return Chroma(
        collection_name=collection_name,
        persist_directory=persist_directory,
        embedding_function=load_embeddings_model(embeddings_model_name),
        client_settings=chroma_settings,
    )


def get_chroma_settings(persist_directory):
    return Settings(
        chroma_db_impl="duckdb+parquet",
        persist_directory=persist_directory,
        anonymized_telemetry=False,
    )


def get_retriever(embeddings_model_name, persist_directory, collection_name):
    chroma_settings = get_chroma_settings(persist_directory)
    db = load_chroma(
        embeddings_model_name, persist_directory, collection_name, chroma_settings
    )
    retriever = db.as_retriever()
    return retriever



from functools import partial

import pandas as pd

from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb_sql.parser.ast import BinaryOperation, Select, Identifier, Constant

from mindsdb.api.mysql.mysql_proxy.utilities.sql import query_df
from mindsdb.api.mysql.mysql_proxy.classes.sql_query import get_all_tables
from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.datanode import DataNode
from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.integration_datanode import IntegrationDataNode
from mindsdb.api.mysql.mysql_proxy.datahub.datanodes.project_datanode import ProjectDataNode
from mindsdb.api.mysql.mysql_proxy.datahub.classes.tables_row import TablesRow, TABLES_ROW_TYPE
from mindsdb.api.mysql.mysql_proxy.utilities import exceptions as exc
from mindsdb.interfaces.database.projects import ProjectController
from mindsdb.interfaces.jobs.jobs_controller import JobsController


class InformationSchemaDataNode(DataNode):
    type = 'INFORMATION_SCHEMA'

    information_schema = {
        'SCHEMATA': ['CATALOG_NAME', 'SCHEMA_NAME', 'DEFAULT_CHARACTER_SET_NAME', 'DEFAULT_COLLATION_NAME', 'SQL_PATH'],
        'TABLES': ['TABLE_CATALOG', 'TABLE_SCHEMA', 'TABLE_NAME', 'TABLE_TYPE', 'ENGINE', 'VERSION', 'ROW_FORMAT', 'TABLE_ROWS', 'AVG_ROW_LENGTH', 'DATA_LENGTH', 'MAX_DATA_LENGTH', 'INDEX_LENGTH', 'DATA_FREE', 'AUTO_INCREMENT', 'CREATE_TIME', 'UPDATE_TIME', 'CHECK_TIME', 'TABLE_COLLATION', 'CHECKSUM', 'CREATE_OPTIONS', 'TABLE_COMMENT'],
        'COLUMNS': ['TABLE_CATALOG', 'TABLE_SCHEMA', 'TABLE_NAME', 'COLUMN_NAME', 'ORDINAL_POSITION', 'COLUMN_DEFAULT', 'IS_NULLABLE', 'DATA_TYPE', 'CHARACTER_MAXIMUM_LENGTH', 'CHARACTER_OCTET_LENGTH', 'NUMERIC_PRECISION', 'NUMERIC_SCALE', 'DATETIME_PRECISION', 'CHARACTER_SET_NAME', 'COLLATION_NAME', 'COLUMN_TYPE', 'COLUMN_KEY', 'EXTRA', 'PRIVILEGES', 'COLUMN_COMMENT', 'GENERATION_EXPRESSION'],
        'EVENTS': ['EVENT_CATALOG', 'EVENT_SCHEMA', 'EVENT_NAME', 'DEFINER', 'TIME_ZONE', 'EVENT_BODY', 'EVENT_DEFINITION', 'EVENT_TYPE', 'EXECUTE_AT', 'INTERVAL_VALUE', 'INTERVAL_FIELD', 'SQL_MODE', 'STARTS', 'ENDS', 'STATUS', 'ON_COMPLETION', 'CREATED', 'LAST_ALTERED', 'LAST_EXECUTED', 'EVENT_COMMENT', 'ORIGINATOR', 'CHARACTER_SET_CLIENT', 'COLLATION_CONNECTION', 'DATABASE_COLLATION'],
        'ROUTINES': ['SPECIFIC_NAME', 'ROUTINE_CATALOG', 'ROUTINE_SCHEMA', 'ROUTINE_NAME', 'ROUTINE_TYPE', 'DATA_TYPE', 'CHARACTER_MAXIMUM_LENGTH', 'CHARACTER_OCTET_LENGTH', 'NUMERIC_PRECISION', 'NUMERIC_SCALE', 'DATETIME_PRECISION', 'CHARACTER_SET_NAME', 'COLLATION_NAME', 'DTD_IDENTIFIER', 'ROUTINE_BODY', 'ROUTINE_DEFINITION', 'EXTERNAL_NAME', 'EXTERNAL_LANGUAGE', 'PARAMETER_STYLE', 'IS_DETERMINISTIC', 'SQL_DATA_ACCESS', 'SQL_PATH', 'SECURITY_TYPE', 'CREATED', 'LAST_ALTERED', 'SQL_MODE', 'ROUTINE_COMMENT', 'DEFINER', 'CHARACTER_SET_CLIENT', 'COLLATION_CONNECTION', 'DATABASE_COLLATION'],
        'TRIGGERS': ['TRIGGER_CATALOG', 'TRIGGER_SCHEMA', 'TRIGGER_NAME', 'EVENT_MANIPULATION', 'EVENT_OBJECT_CATALOG', 'EVENT_OBJECT_SCHEMA', 'EVENT_OBJECT_TABLE', 'ACTION_ORDER', 'ACTION_CONDITION', 'ACTION_STATEMENT', 'ACTION_ORIENTATION', 'ACTION_TIMING', 'ACTION_REFERENCE_OLD_TABLE', 'ACTION_REFERENCE_NEW_TABLE', 'ACTION_REFERENCE_OLD_ROW', 'ACTION_REFERENCE_NEW_ROW', 'CREATED', 'SQL_MODE', 'DEFINER', 'CHARACTER_SET_CLIENT', 'COLLATION_CONNECTION', 'DATABASE_COLLATION'],
        'PLUGINS': ['PLUGIN_NAME', 'PLUGIN_VERSION', 'PLUGIN_STATUS', 'PLUGIN_TYPE', 'PLUGIN_TYPE_VERSION', 'PLUGIN_LIBRARY', 'PLUGIN_LIBRARY_VERSION', 'PLUGIN_AUTHOR', 'PLUGIN_DESCRIPTION', 'PLUGIN_LICENSE', 'LOAD_OPTION', 'PLUGIN_MATURITY', 'PLUGIN_AUTH_VERSION'],
        'ENGINES': ['ENGINE', 'SUPPORT', 'COMMENT', 'TRANSACTIONS', 'XA', 'SAVEPOINTS'],
        'KEY_COLUMN_USAGE': ['CONSTRAINT_CATALOG', 'CONSTRAINT_SCHEMA', 'CONSTRAINT_NAME', 'TABLE_CATALOG', 'TABLE_SCHEMA', 'TABLE_NAME', 'COLUMN_NAME', 'ORDINAL_POSITION', 'POSITION_IN_UNIQUE_CONSTRAINT', 'REFERENCED_TABLE_SCHEMA', 'REFERENCED_TABLE_NAME', 'REFERENCED_COLUMN_NAME'],
        'STATISTICS': ['TABLE_CATALOG', 'TABLE_SCHEMA', 'TABLE_NAME', 'NON_UNIQUE', 'INDEX_SCHEMA', 'INDEX_NAME', 'SEQ_IN_INDEX', 'COLUMN_NAME', 'COLLATION', 'CARDINALITY', 'SUB_PART', 'PACKED', 'NULLABLE', 'INDEX_TYPE', 'COMMENT', 'INDEX_COMMENT', 'IS_VISIBLE', 'EXPRESSION'],
        'CHARACTER_SETS': ['CHARACTER_SET_NAME', 'DEFAULT_COLLATE_NAME', 'DESCRIPTION', 'MAXLEN'],
        'COLLATIONS': ['COLLATION_NAME', 'CHARACTER_SET_NAME', 'ID', 'IS_DEFAULT', 'IS_COMPILED', 'SORTLEN', 'PAD_ATTRIBUTE'],
        # MindsDB specific:
        'MODELS': ['NAME', 'ENGINE', 'PROJECT', 'VERSION', 'STATUS', 'ACCURACY', 'PREDICT', 'UPDATE_STATUS', 'MINDSDB_VERSION', 'ERROR', 'SELECT_DATA_QUERY', 'TRAINING_OPTIONS', 'CURRENT_TRAINING_PHASE', 'TOTAL_TRAINING_PHASES', 'TRAINING_PHASE_NAME', 'TAG', 'CREATED_AT', 'TRAINING_TIME'],
        'MODELS_VERSIONS': ['NAME', 'ENGINE', 'PROJECT', 'ACTIVE', 'VERSION', 'STATUS', 'ACCURACY', 'PREDICT', 'UPDATE_STATUS', 'MINDSDB_VERSION', 'ERROR', 'SELECT_DATA_QUERY', 'TRAINING_OPTIONS', 'TAG', 'CREATED_AT', 'TRAINING_TIME'],
        'DATABASES': ['NAME', 'TYPE', 'ENGINE', 'CONNECTION_DATA'],
        'ML_ENGINES': ['NAME', 'HANDLER', 'CONNECTION_DATA'],
        'HANDLERS': ['NAME', 'TYPE', 'TITLE', 'DESCRIPTION', 'VERSION', 'CONNECTION_ARGS', 'IMPORT_SUCCESS', 'IMPORT_ERROR'],
        'JOBS': ['NAME', 'PROJECT', 'START_AT', 'END_AT', 'NEXT_RUN_AT', 'SCHEDULE_STR', 'QUERY'],
        'JOBS_HISTORY': ['NAME', 'PROJECT', 'RUN_START', 'RUN_END', 'ERROR', 'QUERY']
    }

    def __init__(self, session):
        self.session = session
        self.integration_controller = session.integration_controller
        self.project_controller = ProjectController()
        self.database_controller = session.database_controller

        self.persis_datanodes = {}

        databases = self.database_controller.get_dict()
        if 'files' in databases:
            self.persis_datanodes['files'] = IntegrationDataNode(
                'files',
                ds_type='file',
                integration_controller=self.session.integration_controller
            )

        self.get_dataframe_funcs = {
            'TABLES': self._get_tables,
            'COLUMNS': self._get_columns,
            'SCHEMATA': self._get_schemata,
            'ENGINES': self._get_engines,
            'CHARACTER_SETS': self._get_charsets,
            'COLLATIONS': self._get_collations,
            'MODELS': self._get_models,
            'MODELS_VERSIONS': self._get_models_versions,
            'DATABASES': self._get_databases,
            'ML_ENGINES': self._get_ml_engines,
            'HANDLERS': self._get_handlers,
            'JOBS': self._get_jobs,
            'JOBS_HISTORY': self._get_jobs_history,
        }
        for table_name in self.information_schema:
            if table_name not in self.get_dataframe_funcs:
                self.get_dataframe_funcs[table_name] = partial(self._get_empty_table, table_name)

    def __getitem__(self, key):
        return self.get(key)

    def get(self, name):
        name_lower = name.lower()

        if name.lower() == 'information_schema':
            return self

        if name_lower in self.persis_datanodes:
            return self.persis_datanodes[name_lower]

        existing_databases_meta = self.database_controller.get_dict()  # filter_type='project'
        database_name = None
        for key in existing_databases_meta:
            if key.lower() == name_lower:
                database_name = key
                break

        if database_name is None:
            return None

        database_meta = existing_databases_meta[database_name]
        if database_meta['type'] == 'integration':
            integration = self.integration_controller.get(name=database_name)
            return IntegrationDataNode(
                database_name,
                ds_type=integration['engine'],
                integration_controller=self.session.integration_controller
            )
        if database_meta['type'] == 'project':
            project = self.database_controller.get_project(name=database_name)
            return ProjectDataNode(
                project=project,
                integration_controller=self.session.integration_controller,
                information_schema=self
            )

        integration_names = self.integration_controller.get_all().keys()
        for integration_name in integration_names:
            if integration_name.lower() == name_lower:
                datasource = self.integration_controller.get(name=integration_name)
                return IntegrationDataNode(
                    integration_name,
                    ds_type=datasource['engine'],
                    integration_controller=self.session.integration_controller
                )

        return None

    def has_table(self, tableName):
        tn = tableName.upper()
        if tn in self.information_schema:
            return True
        return False

    def get_table_columns(self, tableName):
        tn = tableName.upper()
        if tn in self.information_schema:
            return self.information_schema[tn]
        raise exc.ErTableExistError(f'Table information_schema.{tableName} does not exists')

    def get_integrations_names(self):
        integration_names = self.integration_controller.get_all().keys()
        # remove files from list to prevent doubling in 'select from INFORMATION_SCHEMA.TABLES'
        return [
            x.lower()
            for x in integration_names
            if x not in ('files', )
        ]

    def get_projects_names(self):
        projects = self.database_controller.get_dict(filter_type='project')
        return [x.lower() for x in projects]

    def _get_handlers(self, query: ASTNode = None):
        columns = self.information_schema['HANDLERS']

        handlers = self.integration_controller.get_handlers_import_status()

        data = []
        for _key, val in handlers.items():
            connection_args = val.get('connection_args')
            if connection_args is not None:
                connection_args = str(dict(connection_args))
            import_success = val.get('import', {}).get('success')
            import_error = val.get('import', {}).get('error_message')
            data.append([
                val['name'], val.get('type'), val.get('title'), val.get('description'), val.get('version'),
                connection_args, import_success, import_error
            ])

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_ml_engines(self, query: ASTNode = None):
        columns = self.information_schema['ML_ENGINES']

        integrations = self.integration_controller.get_all()
        ml_integrations = {
            key: val for key, val in integrations.items()
            if val['type'] == 'ml'
        }

        data = []
        for _key, val in ml_integrations.items():
            data.append([
                val['name'], val.get('engine'), val.get('connection_data')
            ])

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_tables(self, query: ASTNode = None):
        columns = self.information_schema['TABLES']

        target_table = None
        if type(query) == Select and type(query.where) == BinaryOperation and query.where.op == 'and':
            for arg in query.where.args:
                if (
                    type(arg) == BinaryOperation and arg.op == '='
                    and type(arg.args[0]) == Identifier
                    and arg.args[0].parts[-1].upper() == 'TABLE_SCHEMA'
                    and type(arg.args[1]) == Constant
                ):
                    target_table = arg.args[1].value
                    break

        data = []
        for name in self.information_schema.keys():
            if target_table is not None and target_table != name:
                continue
            row = TablesRow(TABLE_TYPE=TABLES_ROW_TYPE.SYSTEM_VIEW, TABLE_NAME=name)
            data.append(row.to_list())

        for ds_name, ds in self.persis_datanodes.items():
            if target_table is not None and target_table != ds_name:
                continue
            ds_tables = ds.get_tables()
            if len(ds_tables) == 0:
                continue
            elif isinstance(ds_tables[0], dict):
                ds_tables = [TablesRow(TABLE_TYPE=TABLES_ROW_TYPE.BASE_TABLE, TABLE_NAME=x['name']) for x in ds_tables]
            elif isinstance(ds_tables, list) and len(ds_tables) > 0 and isinstance(ds_tables[0], str):
                ds_tables = [TablesRow(TABLE_TYPE=TABLES_ROW_TYPE.BASE_TABLE, TABLE_NAME=x) for x in ds_tables]
            for row in ds_tables:
                row.TABLE_SCHEMA = ds_name
                data.append(row.to_list())

        for ds_name in self.get_integrations_names():
            if target_table is not None and target_table != ds_name:
                continue
            try:
                ds = self.get(ds_name)
                ds_tables = ds.get_tables()
                for row in ds_tables:
                    row.TABLE_SCHEMA = ds_name
                    data.append(row.to_list())
            except Exception:
                print(f"Can't get tables from '{ds_name}'")

        for project_name in self.get_projects_names():
            if target_table is not None and target_table != project_name:
                continue
            project_dn = self.get(project_name)
            project_tables = project_dn.get_tables()
            for row in project_tables:
                row.TABLE_SCHEMA = project_name
                data.append(row.to_list())

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_jobs(self, query: ASTNode = None):
        jobs_controller = JobsController()

        project_name = None
        if (
                isinstance(query, Select)
                and type(query.where) == BinaryOperation
                and query.where.op == '='
                and query.where.args[0].parts == ['project']
                and isinstance(query.where.args[1], Constant)
        ):
            project_name = query.where.args[1].value

        data = jobs_controller.get_list(project_name)

        columns = self.information_schema['JOBS']
        columns_lower = [col.lower() for col in columns]

        # to list of lists
        data = [
            [
                row[k]
                for k in columns_lower
            ]
            for row in data
        ]

        return pd.DataFrame(data, columns=columns)

    def _get_jobs_history(self, query: ASTNode = None):
        jobs_controller = JobsController()

        project_name = None
        if (
                isinstance(query, Select)
                and type(query.where) == BinaryOperation
                and query.where.op == '='
                and query.where.args[0].parts == ['project']
                and isinstance(query.where.args[1], Constant)
        ):
            project_name = query.where.args[1].value

        data = jobs_controller.get_history(project_name)

        columns = self.information_schema['JOBS_HISTORY']
        columns_lower = [col.lower() for col in columns]

        # to list of lists
        data = [
            [
                row[k]
                for k in columns_lower
            ]
            for row in data
        ]

        return pd.DataFrame(data, columns=columns)

    def _get_databases(self, query: ASTNode = None):
        columns = self.information_schema['DATABASES']

        project = self.database_controller.get_list()
        data = [
            [x['name'], x['type'], x['engine'], str(x.get('connection_data'))]
            for x in project
        ]

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_models(self, query: ASTNode = None):
        columns = self.information_schema['MODELS']
        data = []
        for project_name in self.get_projects_names():
            project = self.database_controller.get_project(name=project_name)
            project_models = project.get_models()
            for row in project_models:
                table_name = row['name']
                table_meta = row['metadata']
                if table_meta['active'] is not True:
                    continue
                data.append([
                    table_name, table_meta['engine'], project_name, table_meta['version'], table_meta['status'],
                    table_meta['accuracy'], table_meta['predict'], table_meta['update_status'],
                    table_meta['mindsdb_version'], table_meta['error'], table_meta['select_data_query'],
                    table_meta['training_options'], table_meta['current_training_phase'],
                    table_meta['total_training_phases'], table_meta['training_phase_name'],
                    table_meta['label'], row['created_at'], table_meta['training_time']
                ])
            # TODO optimise here
            # if target_table is not None and target_table != project_name:
            #     continue

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_models_versions(self, query: ASTNode = None):
        columns = self.information_schema['MODELS_VERSIONS']
        data = []
        for project_name in self.get_projects_names():
            project = self.database_controller.get_project(name=project_name)
            project_models = project.get_models(active=None)
            for row in project_models:
                table_name = row['name']
                table_meta = row['metadata']
                data.append([
                    table_name, table_meta['engine'], project_name, table_meta['active'], table_meta['version'], table_meta['status'],
                    table_meta['accuracy'], table_meta['predict'], table_meta['update_status'],
                    table_meta['mindsdb_version'], table_meta['error'], table_meta['select_data_query'],
                    table_meta['training_options'], table_meta['label'], row['created_at'], table_meta['training_time']
                ])

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_columns(self, query: ASTNode = None):
        columns = self.information_schema['COLUMNS']

        # NOTE there is a lot of types in mysql, but listed below should be enough for our purposes
        row_templates = {
            'text': ['def', 'SCHEMA_NAME', 'TABLE_NAME', 'COLUMN_NAME', 'COL_INDEX', None, 'YES', 'varchar', 1024, 3072, None, None, None, 'utf8', 'utf8_bin', 'varchar(1024)', None, None, 'select', None, None],
            'timestamp': ['def', 'SCHEMA_NAME', 'TABLE_NAME', 'COLUMN_NAME', 'COL_INDEX', 'CURRENT_TIMESTAMP', 'YES', 'timestamp', None, None, None, None, 0, None, None, 'timestamp', None, None, 'select', None, None],
            'bigint': ['def', 'SCHEMA_NAME', 'TABLE_NAME', 'COLUMN_NAME', 'COL_INDEX', None, 'YES', 'bigint', None, None, 20, 0, None, None, None, 'bigint unsigned', None, None, 'select', None, None],
            'float': ['def', 'SCHEMA_NAME', 'TABLE_NAME', 'COLUMN_NAME', 'COL_INDEX', None, 'YES', 'float', None, None, 12, 0, None, None, None, 'float', None, None, 'select', None, None]
        }

        result = []

        for table_name in self.information_schema:
            table_columns = self.information_schema[table_name]
            for i, column_name in enumerate(table_columns):
                result_row = row_templates['text'].copy()
                result_row[1] = 'information_schema'
                result_row[2] = table_name
                result_row[3] = column_name
                result_row[4] = i
                result.append(result_row)

        mindsdb_dn = self.get('MINDSDB')
        for table_row in mindsdb_dn.get_tables():
            table_name = table_row.TABLE_NAME
            table_columns = mindsdb_dn.get_table_columns(table_name)
            for i, column_name in enumerate(table_columns):
                result_row = row_templates['text'].copy()
                result_row[1] = 'mindsdb'
                result_row[2] = table_name
                result_row[3] = column_name
                result_row[4] = i
                result.append(result_row)

        files_dn = self.get('FILES')
        for table_name in files_dn.get_tables():
            table_columns = files_dn.get_table_columns(table_name)
            for i, column_name in enumerate(table_columns):
                result_row = row_templates['text'].copy()
                result_row[1] = 'files'
                result_row[2] = table_name
                result_row[3] = column_name
                result_row[4] = i
                result.append(result_row)

        df = pd.DataFrame(result, columns=columns)
        return df

    def _get_schemata(self, query: ASTNode = None):
        columns = self.information_schema['SCHEMATA']

        databases_meta = self.session.database_controller.get_list()
        data = [
            ['def', x['name'], 'utf8mb4', 'utf8mb4_0900_ai_ci', None]
            for x in databases_meta
        ]

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_engines(self, query: ASTNode = None):
        columns = self.information_schema['ENGINES']
        data = [['InnoDB', 'DEFAULT', 'Supports transactions, row-level locking, and foreign keys', 'YES', 'YES', 'YES']]

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_charsets(self, query: ASTNode = None):
        columns = self.information_schema['CHARACTER_SETS']
        data = [
            ['utf8', 'UTF-8 Unicode', 'utf8_general_ci', 3],
            ['latin1', 'cp1252 West European', 'latin1_swedish_ci', 1],
            ['utf8mb4', 'UTF-8 Unicode', 'utf8mb4_general_ci', 4]
        ]

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_collations(self, query: ASTNode = None):
        columns = self.information_schema['COLLATIONS']
        data = [
            ['utf8_general_ci', 'utf8', 33, 'Yes', 'Yes', 1, 'PAD SPACE'],
            ['latin1_swedish_ci', 'latin1', 8, 'Yes', 'Yes', 1, 'PAD SPACE']
        ]

        df = pd.DataFrame(data, columns=columns)
        return df

    def _get_empty_table(self, table_name, query: ASTNode = None):
        columns = self.information_schema[table_name]
        data = []

        df = pd.DataFrame(data, columns=columns)
        return df

    def query(self, query: ASTNode, session=None):
        query_tables = get_all_tables(query)

        if len(query_tables) != 1:
            raise exc.ErBadTableError(f'Only one table can be used in query to information_schema: {query}')

        table_name = query_tables[0].upper()

        if table_name not in self.get_dataframe_funcs:
            raise exc.ErNotSupportedYet('Information schema: Not implemented.')

        dataframe = self.get_dataframe_funcs[table_name](query=query)
        data = query_df(dataframe, query, session=self.session)

        columns_info = [
            {
                'name': k,
                'type': v
            }
            for k, v in data.dtypes.items()
        ]

        return data.to_dict(orient='records'), columns_info



class TABLES_ROW_TYPE:
    __slots__ = ()
    BASE_TABLE = 'BASE TABLE'
    VIEW = 'VIEW'
    SYSTEM_VIEW = 'SYSTEM VIEW'


TABLES_ROW_TYPE = TABLES_ROW_TYPE()


class TablesRow:
    columns = [
        'TABLE_CATALOG', 'TABLE_SCHEMA', 'TABLE_NAME', 'TABLE_TYPE',
        'ENGINE', 'VERSION', 'ROW_FORMAT', 'TABLE_ROWS', 'AVG_ROW_LENGTH',
        'DATA_LENGTH', 'MAX_DATA_LENGTH', 'INDEX_LENGTH', 'DATA_FREE',
        'AUTO_INCREMENT', 'CREATE_TIME', 'UPDATE_TIME', 'CHECK_TIME',
        'TABLE_COLLATION', 'CHECKSUM', 'CREATE_OPTIONS', 'TABLE_COMMENT'
    ]

    def __init__(self, TABLE_CATALOG='def', TABLE_SCHEMA='information_schema',
                 TABLE_NAME=None, TABLE_TYPE=TABLES_ROW_TYPE.BASE_TABLE, ENGINE=None,
                 VERSION=None, ROW_FORMAT=None, TABLE_ROWS=0, AVG_ROW_LENGTH=0,
                 DATA_LENGTH=0, MAX_DATA_LENGTH=0, INDEX_LENGTH=0, DATA_FREE=0,
                 AUTO_INCREMENT=None, CREATE_TIME='2022-01-01 00:00:00.0',
                 UPDATE_TIME=None, CHECK_TIME=None, TABLE_COLLATION=None,
                 CHECKSUM=None, CREATE_OPTIONS=None, TABLE_COMMENT=None):
        self.TABLE_CATALOG = TABLE_CATALOG
        self.TABLE_SCHEMA = TABLE_SCHEMA
        self.TABLE_NAME = TABLE_NAME
        self.TABLE_TYPE = TABLE_TYPE
        self.ENGINE = ENGINE
        self.VERSION = VERSION
        self.ROW_FORMAT = ROW_FORMAT
        self.TABLE_ROWS = TABLE_ROWS
        self.AVG_ROW_LENGTH = AVG_ROW_LENGTH
        self.DATA_LENGTH = DATA_LENGTH
        self.MAX_DATA_LENGTH = MAX_DATA_LENGTH
        self.INDEX_LENGTH = INDEX_LENGTH
        self.DATA_FREE = DATA_FREE
        self.AUTO_INCREMENT = AUTO_INCREMENT
        self.CREATE_TIME = CREATE_TIME
        self.UPDATE_TIME = UPDATE_TIME
        self.CHECK_TIME = CHECK_TIME
        self.TABLE_COLLATION = TABLE_COLLATION
        self.CHECKSUM = CHECKSUM
        self.CREATE_OPTIONS = CREATE_OPTIONS
        self.TABLE_COMMENT = TABLE_COMMENT

    def to_list(self) -> list:
        return [self.TABLE_CATALOG, self.TABLE_SCHEMA, self.TABLE_NAME,
                self.TABLE_TYPE, self.ENGINE, self.VERSION, self.ROW_FORMAT,
                self.TABLE_ROWS, self.AVG_ROW_LENGTH, self.DATA_LENGTH,
                self.MAX_DATA_LENGTH, self.INDEX_LENGTH, self.DATA_FREE,
                self.AUTO_INCREMENT, self.CREATE_TIME, self.UPDATE_TIME,
                self.CHECK_TIME, self.TABLE_COLLATION, self.CHECKSUM,
                self.CREATE_OPTIONS, self.TABLE_COMMENT]

    @staticmethod
    def from_dict(data: dict):

        del_keys = []
        data = {k.upper(): v for k, v in data.items()}
        for key in data:
            if key not in TablesRow.columns:
                del_keys.append(key)

        for key in del_keys:
            del data[key]

        return TablesRow(**data)



from mindsdb.api.postgres.postgres_proxy.postgres_proxy import PostgresProxyHandler
from mindsdb.utilities.config import Config
import mindsdb.interfaces.storage.db as db
from mindsdb.utilities import log


def start(verbose=False):
    db.init()

    config = Config()

    log.initialize_log(config, 'postgres_proxy', wrap_print=True)

    PostgresProxyHandler.startProxy()



from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
from collections import OrderedDict


from ..postgres_handler import Handler as PostgresHandler


class MaterializeHandler(PostgresHandler):
    """
    This handler handles connection and execution of the Materialize statements.
    """
    name = 'materialize'

    def __init__(self, name, **kwargs):
        super().__init__(name, **kwargs)






connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the Materialize server/database.'
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Materialize server.'
    },
    password={
        'type': ARG_TYPE.STR,
        'description': 'The password to authenticate the user with the Materialize server.'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'Specify port to connect Materialize server'
    }, 
    database={
        'type': ARG_TYPE.STR,
        'description': 'Specify database name  to connect Materialize server'
    },


)

connection_args_example = OrderedDict(
    host='127.0.0.1',
    port=6875,
    password='',
    user='USER',
    database='materialize'
    

)



from collections import OrderedDict
from typing import Optional
from mindsdb_sql.parser.ast.base import ASTNode
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.utilities import log
from mindsdb_sql import parse_sql
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE



import pandas as pd
import pysqream as db

from pysqream_sqlalchemy.dialect import SqreamDialect 




class SQreamDBHandler(DatabaseHandler):


    name= 'sqreamdb'

    def __init__(self, name: str, connection_data: Optional[dict], **kwargs):
        """ Initialize the handler
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)
        
        
        self.connection_data = connection_data

        self.connection = None
        self.is_connected = False
          
 
    def connect(self):
        """
        Handles the connection to a YugabyteSQL database insance.
        """
        if self.is_connected is True:
            return self.connection
        
        args={
            "database": self.connection_data.get('database'),
            "host": self.connection_data.get('host'),
            "port": self.connection_data.get('port'),
            "username": self.connection_data.get('user'),
            "password": self.connection_data.get('password'),
            "clustered":self.connection_data.get('clustered',False),
            "use_ssl":self.connection_data.get('use_ssl',False), 
            "service":self.connection_data.get('service','sqream')
        }



        connection = db.connect(**args)

        self.is_connected = True
        self.connection = connection
        return self.connection

    def check_connection(self) -> StatusResponse:
        """
        Check the connection of the SQreamDB  database
        :return: success status and error message if error occurs
        """
        response = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            connection = self.connect()
            with connection.cursor() as cur:
                cur.execute('select 1;')
            response.success = True
        except db.Error as e:
            log.logger.error(f'Error connecting to SQreamDB {self.database}, {e}!')
            response.error_message = e

        if response.success is True and need_to_close:
            self.disconnect()
        if response.success is False and self.is_connected is True:
            self.is_connected = False

        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive raw query and act upon it somehow.
        Args:
            query (Any): query in native format (str for sql databases,
                dict for mongo, etc)
        Returns:
            HandlerResponse
        """
        need_to_close = self.is_connected is False
        conn = self.connect()
        with conn.cursor() as cur:
            try:
                cur.execute(query)
                   
                if cur.rowcount >0 and query.upper().startswith('SELECT') :
                    result = cur.fetchall() 
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        data_frame=pd.DataFrame(
                            result,
                            columns=[x[0] for x in cur.description]
                        )
                    )
                else:
                    response = Response(RESPONSE_TYPE.OK)
                self.connection.commit()
            except Exception as e:
                log.logger.error(f'Error running query: {query} on {self.database}!')
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_message=str(e)
                )
                self.connection.rollback()

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Retrieve the data from the SQL statement
        """
        renderer = SqlalchemyRender(SqreamDialect)
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        List all tables in SQreamDB stored in 'sqream_catalog' 
        """

        query = "SELECT table_name FROM sqream_catalog.tables"

        return self.query(query)

    def get_columns(self, table_name):
        query = f"""SELECT column_name, type_name 
        FROM sqream_catalog.columns 
        WHERE table_name = '{table_name}';
        """
        return self.query(query)



connection_args = OrderedDict(
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the SQreamDB server/database.'
    },
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the SQreamDB server.'
    },
    password={
        'type': ARG_TYPE.STR,
        'description': 'The password to authenticate the user with the SQreamDB server.'
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'Specify port to connect SQreamDB server'
    }, 
    database={
        'type': ARG_TYPE.STR,
        'description': 'Specify database name  to connect SQreamDB server'
    },
    service={
        'type': ARG_TYPE.STR,
        'description': 'Optional: service queue (default: "sqream")'
    },
    use_ssl={
        'type': ARG_TYPE.BOOL,
        'description': 'use SSL connection (default: False)'
    },
    clustered={
        'type': ARG_TYPE.BOOL,
        'description': 'Optional:  Connect through load balancer, or direct to worker (Default: false - direct to worker)'
    },


)

connection_args_example = OrderedDict(
    host='127.0.0.1',
    port=5000,
    password='sqream',
    user='master',
    database='sqream'
    

)



import requests
from bs4 import BeautifulSoup
from urllib.parse import urlparse, urljoin
import logging
import re
import traceback
import pandas as pd
from threading import Lock

import requests
import concurrent.futures
from urllib.parse import urlparse


url_list_lock = Lock()

def is_valid(url):
    parsed = urlparse(url)
    return bool(parsed.netloc) and bool(parsed.scheme)


# this bad boy gets all the crawling done in parallel
def parallel_get_all_website_links(urls):
    url_contents = {}

    if len(urls) <= 10:
        for url in urls:
            url_contents[url] = get_all_website_links(url)
        return url_contents
    
    with concurrent.futures.ProcessPoolExecutor() as executor:
        future_to_url = {executor.submit(get_all_website_links, url): url for url in urls}
        for future in concurrent.futures.as_completed(future_to_url):
            url = future_to_url[future]
            try:
                url_contents[url] = future.result()
            except Exception as exc:
                logging.error(f'{url} generated an exception: {exc}')
   
    return url_contents


# this crawls one individual website
def get_all_website_links(url):
    logging.info('crawling: {url} ...'.format(url=url))
    urls = set()
            
    domain_name = urlparse(url).netloc
    try:
        content_html = requests.get(url).content
        soup = BeautifulSoup(content_html, "html.parser")
    except Exception as e:
        error_message = traceback.format_exc().splitlines()[-1]
        logging.error("An exception occurred: %s", str(e))
        return {'url':url,'urls':urls, 'html_content':'', 'text_content': '', 'error':str(error_message)}
    
    content_text = get_readable_text_from_soup(soup)
    for a_tag in soup.findAll("a"):
        href = a_tag.attrs.get("href")
        if href == "" or href is None:
            continue
        href = urljoin(url, href)
        parsed_href = urlparse(href)
        
        href = parsed_href.scheme + "://" + parsed_href.netloc + parsed_href.path
        if not is_valid(href):
            continue
        if href in urls:
            continue
        if domain_name != parsed_href.netloc:
            continue
        
        href = href.rstrip('/')    
        urls.add(href)

    return {'url': url,'urls': urls, 'html_content': content_html, 'text_content': content_text, 'error': None}


# this returns the soup object
def get_readable_text_from_soup(soup):

    # Remove script and style elements
    for script in soup(["script", "style"]):
        script.decompose()

    # Get text, #todo add new line for li elements
    text = soup.get_text()

    # Remove leading and trailing spaces on each line
    lines = (line.strip() for line in text.splitlines())
    # Break multi-headlines into a line each
    chunks = (phrase.strip() for line in lines for phrase in line.split("  "))
    # Remove blank lines
    text = '\n'.join(chunk for chunk in chunks if chunk)

    return text


# this bad girl does the recursive crawling of the websites
def get_all_website_links_rec(url, reviewd_urls, limit=None):

    if limit is not None:
        if len(reviewd_urls) >= limit:
            return reviewd_urls

    if url not in reviewd_urls:
        # if something happens getting the website links for this url then log the error
        try:
            reviewd_urls[url] = get_all_website_links(url)
        except Exception as e:
            error_message = traceback.format_exc().splitlines()[-1]
            logging.error("An exception occurred: %s", str(e))
            reviewd_urls[url] = {'url': url, 'urls': [], 'html_content': '', 'text_content': '', 'error': str(error_message)}

    to_rev_url_list = []

    # create a list of new urls to review that don't exist in the already reviewed ones
    for new_url in reviewd_urls[url]['urls']:
        
        # if this is already in the urls, then no need to go and crawl for it
        if new_url in reviewd_urls or new_url in to_rev_url_list:
            continue
        
        # insert immediately to count limit between threads. fill later
        with url_list_lock:
            if limit is None or len(reviewd_urls) < limit:
                reviewd_urls[new_url] = {}
                to_rev_url_list.append(new_url)
            else:
                break

    # if there is something to fetch, go fetch
    if len(to_rev_url_list) > 0:
        new_revised_urls = parallel_get_all_website_links(to_rev_url_list)
    
        reviewd_urls.update(new_revised_urls)

        for new_url in new_revised_urls:
            get_all_website_links_rec(new_url, reviewd_urls, limit)


# this crawls the websites and retuns it all as a dataframe, ready to be served
def get_all_websites(urls, limit=1, html=False):
    reviewd_urls = {}

    # def fetch_url(url):
    #     url = url.rstrip('/')
    #     if urlparse(url).scheme == "":
    #         # Try HTTPS first
    #         url = "https://" + url
    #     reviewd_urls_iter = {}
    #     get_all_website_links_rec(url, reviewd_urls_iter, limit)
    #     return reviewd_urls_iter
    
    # reviewd_urls = fetch_url(urls[0])
    # Define a helper function that will be run in parallel.
    def fetch_url(url):
        url = url.rstrip('/')
        if urlparse(url).scheme == "":
            # Try HTTPS first
            url = "https://" + url
        get_all_website_links_rec(url, reviewd_urls, limit)

    # Use a ThreadPoolExecutor to run the helper function in parallel.
    with concurrent.futures.ThreadPoolExecutor() as executor:
        future_to_url = {executor.submit(fetch_url, url): url for url in urls}

        for future in concurrent.futures.as_completed(future_to_url):
            future.result()

    columns_to_ignore = ['urls']
    if html is False:
        columns_to_ignore += ['html_content']
    df = dict_to_dataframe(reviewd_urls, columns_to_ignore=columns_to_ignore, index_name='url')

    return df


# this can parse the native query    
def parse_urls_limit(input_str):
    # Split the input string into 'url', 'limit' or 'html' parts
    items = re.split(r',', input_str)

    # Initialize list for urls, limit and html
    urls = []
    limit = None
    html = False

    for item in items:
        item = item.strip()  # Remove leading/trailing whitespace

        # Check if item is a 'limit' or 'html' setting
        if item.lower().startswith('limit'):
            limit_match = re.search(r'\d+', item)
            if limit_match:
                limit = int(limit_match.group())  # Extract the number
        elif item.lower().startswith('html'):
            html_match = re.search(r'(true|false)', item, re.I)
            if html_match:
                html = html_match.group().lower() == 'true'  # Check if the value is 'true'
        else:
            urls.append(item)  # Add the item to the url list

    return {"urls": urls, "limit": limit, "html": html}


# run a query that goes and crawls urls
# format url, url, ..., limit=n
# you can pass one of many urls, limit is optional
def get_df_from_query_str(query_str):
    args = parse_urls_limit(query_str)
    df = get_all_websites(args['urls'], args['limit'], args['html'])
    return df


# this flips a dictionary of dictionaries into a dataframe so we can use it in mindsdb
def dict_to_dataframe(dict_of_dicts, columns_to_ignore=None, index_name=None):
    # Convert dictionary of dictionaries into DataFrame
    df = pd.DataFrame.from_dict(dict_of_dicts, orient='index')

    # If columns_to_ignore is provided, drop these columns
    if columns_to_ignore:
        df = df.drop(columns_to_ignore, axis=1, errors='ignore')

    # If index_name is provided, rename the index
    if index_name:
        df.index.name = index_name

    return df


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    df = get_df_from_query_str('docs.mindsdb.com, docs.airbyte.com, limit=4')
    print(df)




from mindsdb.integrations.handlers.jira_handler.jira_table import JiraProjectsTable
from mindsdb.integrations.libs.api_handler import APIHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
)
from mindsdb.utilities.log import get_log
from mindsdb_sql import parse_sql

from atlassian import Jira
from typing import Optional
import requests


log = get_log("integrations.jira_handler")


class JiraHandler(APIHandler):
    """
    This handler handles connection and execution of the Airtable statements.
    """


    def __init__(self, name=None, **kwargs):
        """
        Initialize the handler.
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
            **kwargs: arbitrary keyword arguments.
        """
        super().__init__(name)
        self.connection_data = kwargs.get("connection_data", {})

        self.parser = parse_sql
        self.dialect = 'jira'
        self.kwargs = kwargs
        self.connection = None
        self.is_connected = False


        jira_projects_data = JiraProjectsTable(self)
        self._register_table("project", jira_projects_data)

    def __del__(self):
        if self.is_connected is True:
            self.disconnect()

    def connect(self) -> StatusResponse:
        """
        Set up the connection required by the handler.
        Returns:
            HandlerStatusResponse
        """

        if self.is_connected is True:
            return self.connection
        
        s = requests.Session()
        s.headers['Authorization'] =  f"Bearer {self.connection_data['jira_api_token']}"

        self.connection = Jira(url= self.connection_data['jira_url'], session=s)
        self.is_connected = True


        return self.connection

    def disconnect(self):
        """
        Close any existing connections.
        """

        if self.is_connected is False:
            return

        self.connection.close()
        self.is_connected = False
        return self.is_connected

    def check_connection(self) -> StatusResponse:
        """
        Check connection to the handler.
        Returns:
            HandlerStatusResponse
        """

        response = StatusResponse(False)
        need_to_close = self.is_connected is False
        
        try:
            self.connect()
            response.success = True
        except Exception as e:
            logger.error(f"Error connecting to Jira API: {e}!")
            response.error_message = e

        self.is_connected = response.success

        return response

    def native_query(self, query: str) -> StatusResponse:
        """Receive and process a raw query.
        Parameters
        ----------
        query : str
            query in a native format
        Returns
        -------
        StatusResponse
            Request status
        """
        ast = parse_sql(query, dialect="mindsdb")
        return self.query(ast)
   
    def construct_jql(self):
        """construct jql & returns it to JiraProjectsTable class
        Returns
        -------
        Str
        """
        return 'project = ' + str(self.connection_data['project'])



from mindsdb.api.mysql.mysql_proxy.mysql_proxy import MysqlProxy
import mindsdb.interfaces.storage.db as db
from mindsdb.utilities.config import Config
from mindsdb.utilities import log
from mindsdb.utilities.functions import init_lexer_parsers


def start(verbose=False):
    db.init()
    init_lexer_parsers()

    config = Config()

    log.initialize_log(config, 'mysql', wrap_print=True)

    MysqlProxy.startProxy()



import os
from typing import Optional, Dict

import pandas as pd

from hugging_py_face import NLP, ComputerVision, AudioProcessing

from mindsdb.utilities.config import Config
from mindsdb.integrations.libs.base import BaseMLEngine


class HuggingFaceInferenceAPIHandler(BaseMLEngine):
    """
    Integration with the Hugging Face Inference API.
    """

    name = 'huggingface_api'

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        if 'using' not in args:
            raise Exception("Hugging Face Inference engine requires a USING clause! Refer to its documentation for more details.")

        self.model_storage.json_set('args', args)

    def predict(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:
        args = self.model_storage.json_get('args')
        api_key = self._get_huggingface_api_key(args)

        if args['using']['task'] == 'text-classification':
            nlp = NLP(api_key)
            result_df = nlp.text_classification_in_df(
                df,
                args['using']['column'],
                args['using']['options'] if 'options' in args['using'] else None,
                args['using']['model'] if 'model' in args['using'] else None
            )

        elif args['using']['task'] == 'fill-mask':
            nlp = NLP(api_key)
            result_df = nlp.fill_mask_in_df(
                df,
                args['using']['column'],
                args['using']['options'] if 'options' in args['using'] else None,
                args['using']['model'] if 'model' in args['using'] else None
            )

        elif args['using']['task'] == 'summarization':
            nlp = NLP(api_key)
            result_df = nlp.summarization_in_df(
                df,
                args['using']['column'],
                args['using']['parameters'] if 'parameters' in args['using'] else None,
                args['using']['options'] if 'options' in args['using'] else None,
                args['using']['model'] if 'model' in args['using'] else None
            )

        elif args['using']['task'] == 'text-generation':
            nlp = NLP(api_key)
            result_df = nlp.text_generation_in_df(
                df,
                args['using']['column'],
                args['using']['parameters'] if 'parameters' in args['using'] else None,
                args['using']['options'] if 'options' in args['using'] else None,
                args['using']['model'] if 'model' in args['using'] else None
            )

        elif args['using']['task'] == 'question-answering':
            nlp = NLP(api_key)
            result_df = nlp.question_answering_in_df(
                df,
                args['using']['question_column'],
                args['using']['context_column'],
                args['using']['model'] if 'model' in args['using'] else None
            )

        elif args['using']['task'] == 'sentence-similarity':
            nlp = NLP(api_key)
            result_df = nlp.sentence_similarity_in_df(
                df,
                args['using']['source_sentence_column'],
                args['using']['sentence_column'],
                args['using']['options'] if 'options' in args['using'] else None,
                args['using']['model'] if 'model' in args['using'] else None
            )

        elif args['using']['task'] == 'zero-shot-classification':
            nlp = NLP(api_key)
            result_df = nlp.zero_shot_classification_in_df(
                df,
                args['using']['column'],
                args['using']['candidate_labels'],
                args['using']['parameters'] if 'parameters' in args['using'] else None,
                args['using']['options'] if 'options' in args['using'] else None,
                args['using']['model'] if 'model' in args['using'] else None
            )

        elif args['using']['task'] == 'image-classification':
            cp = ComputerVision(api_key)
            result_df = cp.image_classification_in_df(
                df,
                args['using']['column'],
                args['using']['model'] if 'model' in args['using'] else None
            )

        elif args['using']['task'] == 'object-detection':
            cp = ComputerVision(api_key)
            result_df = cp.object_detection_in_df(
                df,
                args['using']['column'],
                args['using']['model'] if 'model' in args['using'] else None
            )

        elif args['using']['task'] == 'speech-recognition':
            ap = AudioProcessing(api_key)
            result_df = ap.speech_recognition_in_df(
                df,
                args['using']['column'],
                args['using']['model'] if 'model' in args['using'] else None
            )

        elif args['using']['task'] == 'audio-classification':
            ap = AudioProcessing(api_key)
            result_df = ap.audio_classification_in_df(
                df,
                args['using']['column'],
                args['using']['model'] if 'model' in args['using'] else None
            )

        else:
            raise Exception(f"Task {args['using']['task']} is not supported!")

        result_df = result_df.rename(columns={'predictions': args['target']})
        return result_df

    def _get_huggingface_api_key(self, args, strict=True):
        """ 
        API_KEY preference order:
            1. provided at model creation
            2. provided at engine creation
            3. HUGGINGFACE_API_KEY env variable
            4. huggingface.api_key setting in config.json
        """  # noqa
        # 1
        if 'api_key' in args:
            return args['api_key']
        # 2
        connection_args = self.engine_storage.get_connection_args()
        if 'api_key' in connection_args:
            return connection_args['api_key']
        # 3
        api_key = os.getenv('HUGGINGFACE_API_KEY')
        if api_key is not None:
            return api_key
        # 4
        config = Config()
        openai_cfg = config.get('huggingface', {})
        if 'api_key' in openai_cfg:
            return openai_cfg['api_key']

        if strict:
            raise Exception(f'Missing API key "api_key". Either re-create this ML_ENGINE specifying the `api_key` parameter,\
                 or re-create this model and pass the API key with `USING` syntax.')  # noqa


import paypalrestsdk
import pandas as pd
from typing import Text, List, Dict

from mindsdb_sql.parser import ast
from mindsdb.integrations.libs.api_handler import APITable

from mindsdb.integrations.handlers.utilities.query_utilities import SELECTQueryParser, SELECTQueryExecutor


class PaymentsTable(APITable):

    def select(self, query: ast.Select) -> pd.DataFrame:
        """
        Pulls PayPal Payments data.
        Parameters
        ----------
        query : ast.Select
           Given SQL SELECT query
        Returns
        -------
        pd.DataFrame
            PayPal Payments matching the query
        Raises
        ------
        ValueError
            If the query contains an unsupported condition
        """

        select_statement_parser = SELECTQueryParser(
            query,
            'payments',
            self.get_columns()
        )
        selected_columns, where_conditions, order_by_conditions, result_limit = select_statement_parser.parse_query()

        payments_df = pd.json_normalize(self.get_payments(count=result_limit))
        select_statement_executor = SELECTQueryExecutor(
            payments_df,
            selected_columns,
            where_conditions,
            order_by_conditions
        )
        payments_df = select_statement_executor.execute_query()

        return payments_df

    def get_columns(self) -> List[Text]:
        return pd.json_normalize(self.get_payments(count=1)).columns.tolist()

    def get_payments(self, **kwargs) -> List[Dict]:
        connection = self.handler.connect()
        payments = paypalrestsdk.Payment.all(kwargs, api=connection)
        return [payment.to_dict() for payment in payments['payments']]



import datetime as dt
from collections import defaultdict

from mindsdb_sql.parser.ast import Identifier, Select, Insert, BinaryOperation, Constant
from mindsdb.interfaces.storage import db

from mindsdb.integrations.libs.api_handler import APIChatHandler


class BotException(Exception):
    pass


class ChatBotTask:
    def __init__(self, session, database, project_name, model_name, params):

        self.session = session
        self.project_name = project_name

        self.db_handler = self.session.integration_controller.get_handler(database)
        self.project_datanode = self.session.datahub.get(project_name)

        # get chat handler info
        self.params = {}
        if isinstance(self.db_handler, APIChatHandler):
            self.params = self.db_handler.get_chat_config()

            # get bot username
            self.params['bot_username'] = self.db_handler.get_my_user_name()

        if params is not None:
            self.params.update(params)

        # previous chats information
        self.chats_prev = None

        # get model info
        self.params['model'] = self._get_model(model_name)

        self.chat_memory = defaultdict(dict)

        self.params['back_db'] = {}
        back_db = self.params.get('backoffice_db')
        if back_db is not None:
            self.back_db = self.session.integration_controller.get_handler(back_db)

            if hasattr(self.back_db, 'back_office_config'):
                self.params['back_db']['config'] = self.back_db.back_office_config()

            if 'tools' in self.params['back_db']['config']:
                tools = [
                    {
                        'name': name,
                        'func': getattr(self.back_db, name),
                        'description': description
                    }
                    for name, description in self.params['back_db']['config']['tools'].items()
                ]
                self.params['back_db']['tools'] = tools

    def _get_model(self, model_name):
        model = self.session.model_controller.get_model(model_name, project_name=self.project_name)
        model_record = db.Predictor.query.get(model['id'])
        integration_record = db.Integration.query.get(model_record.integration_id)

        return {
            'model_name': model_name,
            'user_column': model_record.learn_args['using']['user_column'],
            'bot_column': model_record.learn_args['using']['assistant_column'],
            'output': model_record.to_predict[0],
            'engine': integration_record.engine,
        }

    def run(self):

        # get chat ids to react
        if self.params['polling']['type'] == 'message_count':
            chat_ids = self.get_chats_by_message_count()
        else:
            raise NotImplementedError

        for chat_id in chat_ids:
            self.answer_to_user(chat_id)

    def get_chats_by_message_count(self):

        p_params = self.params['polling']

        chat_ids = []

        id_col = p_params['chat_id_col']
        msgs_col = p_params['count_col']
        # get chats status info
        ast_query = Select(
            targets=[
                Identifier(id_col),
                Identifier(msgs_col)],
            from_table=Identifier(p_params['table'])
        )

        resp = self.db_handler.query(query=ast_query)
        if resp.data_frame is None:
            raise BotException('Error to get count of messages')

        chats = {}
        for row in resp.data_frame.to_dict('records'):
            chat_id = row[id_col]
            msgs = row[msgs_col]

            chats[chat_id] = msgs

        if self.chats_prev is None:
            # first run
            self.chats_prev = chats
        else:
            # compare
            # for new keys
            for chat_id, count_msgs in chats.items():
                if self.chats_prev.get(chat_id) != count_msgs:
                    chat_ids.append(chat_id)

            self.chats_prev = chats
        return chat_ids

    def answer_to_user(self, chat_id):
        bot_username = self.params['bot_username']

        t_params = self.params['chat_table']

        text_col = t_params['text_col']
        username_col = t_params['username_col']
        time_col = t_params['time_col']

        ast_query = Select(
            targets=[Identifier(text_col),
                     Identifier(username_col),
                     Identifier(time_col)],
            from_table=Identifier(t_params['name']),
            where=BinaryOperation(
                op='=',
                args=[
                    Identifier(t_params['chat_id_col']),
                    Constant(chat_id)
                ]
            )
        )

        resp = self.db_handler.query(ast_query)
        if resp.data_frame is None:
            raise Exception('Error to get list of messages')

        df = resp.data_frame

        # check first message:
        if len(df) == 0:
            return
        if df[username_col][0] == bot_username:
            # the last message is from bot
            return

        df = df.rename(columns={
            text_col: 'text',
            username_col: 'username',
            time_col: 'sent_at'
        })
        model_output = self.apply_model(df, chat_id)

        # send answer to user
        ast_query = Insert(
            table=Identifier(t_params['name']),
            columns=[t_params['chat_id_col'], t_params['text_col']],
            values=[
                [chat_id, model_output],
            ]
        )

        self.db_handler.query(ast_query)

    def _chat_history_to_conversation(self, df, chat_id):

        bot_username = self.params['bot_username']

        question_col = self.params['model']['user_column']
        answer_col = self.params['model']['bot_column']

        history_from = self.chat_memory[chat_id].get('history_from')
        if history_from:
            df = df[df.sent_at >= history_from]

        messages = []
        msgs = df.to_dict('records')

        for row in msgs:
            text = row['text']

            if text is None or text.strip() == '':
                # skip empty rows
                continue

            if row['username'] != bot_username:
                # create new message row
                messages.append({question_col: text, answer_col: None})
            else:
                if len(messages) == 0:
                    # add empty row
                    messages.append({question_col: None, answer_col: None})

                # update answer in previous column
                messages[-1][answer_col] = text
        return messages

    def _make_mode_prompt(self, chat_id):
        # choose prompt or model depending on mode
        mode_name = self.chat_memory[chat_id].get('mode')

        prompt = None
        model_info = None
        allowed_tools = None

        if mode_name is None:
            # mode in not selected, lets to go to select menu
            prompt = self._make_select_mode_prompt()
        else:
            # mode is selected

            mode = self.params['modes'].get(mode_name)
            if mode is None:
                # wrong mode
                self.chat_memory[chat_id]['mode'] = None
                raise BotException(f'Error to use mode: {mode_name}')

            if 'model' in mode:
                # this is model
                model_info = self._get_model(mode['model'])

            elif 'prompt' in mode:
                # it is just a prompt. let's use a bot model and custom prompt
                prompt = mode['prompt']

            else:
                raise BotException(f'Mode is not supported: {mode}')

            allowed_tools = mode.get('allowed_tools')

        return prompt, model_info, allowed_tools

    def _get_avail_modes_items(self):
        return [
            f'- code: {key}, description: {value["info"]}'
            for key, value in self.params.get('modes', {}).items()
        ]

    def _make_select_mode_prompt(self):
        # select mode tool
        task_items = self._get_avail_modes_items()

        tasks = '\n'.join(task_items)

        prompt = f'You are a helpful assistant and you can help with various types of tasks.' \
                 f'\nAvailable types of tasks:' \
                 f'\n{tasks}' \
                 f'\nUser have to choose task and assistant MUST call select_task function after it'

        return prompt

    def apply_model(self, df, chat_id):
        # sort by time
        df = df.sort_values(by='sent_at')

        model_info = self.params['model']
        prompt = None
        tools = None
        allowed_tools = None

        switched_to_mode = []

        def _select_task(mode_name):
            if mode_name == '':
                self.chat_memory[chat_id]['mode'] = None
                switched_to_mode.append(None)
                return 'success'

            avail_modes = list(self.params['modes'].keys())
            if mode_name not in avail_modes:
                return f'Error: task is not found. Available tasks: {", ".join(avail_modes)}'
            self.chat_memory[chat_id]['mode'] = mode_name
            switched_to_mode.append(mode_name)
            return 'success'

        select_task_tool = {
            'name': 'select_task',
            'func': _select_task,
            'description': 'Have to be used by assistant to select task. Input is task type. '
                           'If user want to unselect task input should be empty string. '
                           'Available tasks: ' + '; '.join(self._get_avail_modes_items())
        }

        if self.params.get('modes') is not None:
            # a bot with modes
            prompt, model_info2, allowed_tools = self._make_mode_prompt(chat_id)
            if model_info2 is not None:
                model_info = model_info2

            # workaround: don't show history if mode is not selected, otherwise bot doesn't decide to change mode
            if self.chat_memory[chat_id].get('mode') is None:
                df = df[-1:]

        if model_info['engine'] == 'langchain':
            tools = []
            for tool in self.params['back_db']['tools']:
                if allowed_tools is None or tool['name'] in allowed_tools:
                    tools.append(tool)

            if self.params.get('modes') is not None:
                # add mode tool
                tools.append(select_task_tool)

        context_list = [
            f"- Today's date is {dt.datetime.now().strftime('%Y-%m-%d')}."
            f" It must be used to understand the input date from string like 'tomorrow', 'today', 'yesterday'"
        ]
        context = '\n'.join(context_list)

        # call model
        params = {'tools': tools, 'context': context, 'max_iterations': 10, 'prompt': prompt}

        messages = self._chat_history_to_conversation(df, chat_id)

        predictions = self.project_datanode.predict(
            model_name=model_info['model_name'],
            data=messages,
            params=params
        )

        if len(switched_to_mode) > 0:
            # mode changed:
            # - clear previous history
            # - run once again

            # start conversation only from last message
            self.chat_memory[chat_id]['history_from'] = df.iloc[-1]['sent_at']

            messages = self._chat_history_to_conversation(df, chat_id)

            prompt, model_info2, allowed_tools = self._make_mode_prompt(chat_id)

            params['tools'] = []
            for tool in self.params['back_db']['tools']:
                if allowed_tools is None or tool['name'] in allowed_tools:
                    params['tools'].append(tool)

            params['prompt'] = prompt
            if model_info2 is not None:
                model_info = model_info2

            predictions = self.project_datanode.predict(
                model_name=model_info['model_name'],
                data=messages,
                params=params
            )

        output_col = model_info['output']
        model_output = predictions.iloc[-1][output_col]
        return model_output



from mindsdb.integrations.handlers.gmail_handler.gmail_handler import GmailHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE
import unittest


class EmailsTableTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "credentials_file": "mindsdb/integrations/handlers/gmail_handler/credentials.json",
                "scopes": ['https://www.googleapis.com/auth/gmail.readonly']
            }
        }

        cls.handler = GmailHandler('test_gmail_handler', **cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_native_query_select(self):
        query = 'SELECT * FROM emails WHERE query = "alert from:*@google.com"'
        result = self.handler.native_query(query)
        assert result.type is RESPONSE_TYPE.TABLE

    def test_2_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is not RESPONSE_TYPE.ERROR

    def test_3_get_columns(self):
        columns = self.handler.emails.get_columns()

        expected_columns = [
            'id',
            'message_id',
            'thread_id',
            'label_ids',
            'sender',
            'to',
            'date',
            'subject',
            'snippet',
            'body',
        ]

        self.assertListEqual(columns, expected_columns)


if __name__ == '__main__':
    unittest.main()



from typing import Text, List, Dict, Any, Optional

from mindsdb_sql.parser import ast

from .exceptions import UnsupportedColumnException, MandatoryColumnException, ColumnCountMismatchException


class INSERTQueryParser:
    """
    Parses a INSERT query into its component parts.

    Parameters
    ----------
    query : ast.Insert
        Given SQL INSERT query.
    supported_columns : List[Text], Optional
        List of columns supported by the table.
    mandatory_columns : List[Text], Optional
        List of columns that must be present in the query.
    all_mandatory : Optional[Any], Optional (default=True)
        Whether all mandatory columns must be present in the query. If False, only one of the mandatory columns must be present.
    """
    def __init__(self, query: ast.Insert, supported_columns: Optional[List[Text]] = None, mandatory_columns: Optional[List[Text]] = None, all_mandatory: Optional[Any] = True):
        self.query = query
        self.supported_columns = supported_columns
        self.mandatory_columns = mandatory_columns
        self.all_mandatory = all_mandatory

    def parse_query(self) -> List[Dict[Text, Any]]:
        """
        Parses a SQL INSERT statement into its components: columns, values and returns a list of dictionaries with the values to insert.
        """
        columns = self.parse_columns()
        values = self.parse_values()

        values_to_insert = []
        for value in values:
            if len(columns) != len(value):
                raise ColumnCountMismatchException("Number of columns does not match the number of values")
            else:
                values_to_insert.append(dict(zip(columns, value)))

        return values_to_insert

    def parse_columns(self):
        """
        Parses the columns in the query. Raises an exception if the columns are not supported or if mandatory columns are missing.
        """
        columns = [col.name for col in self.query.columns]

        if self.supported_columns:
            if not set(columns).issubset(self.supported_columns):
                unsupported_columns = set(columns).difference(self.supported_columns)
                raise UnsupportedColumnException(f"Unsupported columns: {', '.join(unsupported_columns)}")

        if self.mandatory_columns:
            if self.all_mandatory:
                if not set(self.mandatory_columns).issubset(columns):
                    missing_mandatory_columns = set(self.mandatory_columns).difference(columns)
                    raise MandatoryColumnException(f"Mandatory columns missing: {', '.join(missing_mandatory_columns)}")
            else:
                if not set(self.mandatory_columns).intersection(columns):
                    missing_mandatory_columns = set(self.mandatory_columns).difference(columns)
                    raise MandatoryColumnException(f"Mandatory columns missing: {', '.join(missing_mandatory_columns)}")

        return columns

    def parse_values(self):
        """
        Parses the values in the query.
        """
        return self.query.values


"""
Utility functions for run ml handler in subprocess.
Interation is performed using stdin, stdout.

Flow
    1. subprocess starts and initialize ml hanlder instance inside itself
    2. waits incoming command from stdin
    3. passes this command to ml handler instance
    4. sends response from ml handler to stdout

Protocol over stdin/out:

- [8 bytes] length of the command
- [length] command: is picked dict with keys: method, args, kwargs

if 'error' in response dict, parent process trows exception

""" # noqa

import pickle
import subprocess
import sys
import struct
import traceback
import importlib
import threading
import queue
from threading import Lock


# =================  child process ====================

class MLHandlerProcess:

    def __init__(self):
        self.ml_instance = None

    def mainloop(self):
        # replace print output to stderr
        sys.stdout = sys.stderr

        self.stdin = open(0, 'rb')
        self.stdout = open(1, 'wb')

        while True:
            params = self.read_input()
            try:
                method = params['method']
                args = params.get('args', [])
                kwargs = params.get('kwargs', {})

                if hasattr(self, method):
                    # it is control method
                    ret = getattr(self, method)(*args, **kwargs)
                else:
                    # it is handler method
                    ret = getattr(self.ml_instance, method)(*args, **kwargs)

            except SystemExit:
                raise
            except RuntimeError as e:
                # don't show traceback f
                self.send_output({'protocol_error': str(e)})
                continue
            except Exception as e:
                error = traceback.format_exc()
                self.send_output({'protocol_error': str(e), 'trace': error})
                continue

            self.send_output(ret)

    def send_output(self, obj):
        # read stdin
        encoded = pickle.dumps(obj)

        length_enc = struct.pack('L', len(encoded))

        self.stdout.write(length_enc)
        self.stdout.write(encoded)
        self.stdout.flush()

    def read_input(self):
        # write to stdout
        length_enc = self.stdin.read(8)
        length = struct.unpack('L', length_enc)[0]

        encoded = self.stdin.read(length)
        obj = pickle.loads(encoded)

        return obj

    def init_handler(self, class_path, integration_id, predictor_id, context_dump):
        # mdb initialization
        import mindsdb.interfaces.storage.db as db
        from mindsdb.utilities.context import context as ctx
        ctx.load(context_dump)
        db.init()

        from mindsdb.interfaces.storage.model_fs import ModelStorage, HandlerStorage

        module_name, class_name = class_path
        module = importlib.import_module(module_name)
        HandlerClass = getattr(module, class_name)

        handlerStorage = HandlerStorage(integration_id)
        modelStorage = ModelStorage(predictor_id)

        ml_handler = HandlerClass(
            engine_storage=handlerStorage,
            model_storage=modelStorage,
        )
        self.ml_instance = ml_handler

    def exit(self):
        sys.exit(0)


# =================  parent wrapper ====================

class MLHandlerWrapper:
    def __init__(self):
        # TODO change to virtualenv from config
        python_path = sys.executable
        wrapper_path = __file__
        p = subprocess.Popen(
            [python_path, wrapper_path],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        self.proc = p

    def __del__(self):
        if self.proc is not None:
            self.close()

    def close(self):
        self.send_command({'method': 'exit'})
        # self.proc.stdin.close()
        self.proc.wait()
        self.proc = None

    def __getattr__(self, method_name):
        # is call of the method
        def call(*args, **kwargs):
            return self.exec_command(method_name, *args, **kwargs)
        return call

    def send_command(self, params):
        params_enc = pickle.dumps(params)
        length_enc = struct.pack('L', len(params_enc))

        # send len
        self.proc.stdin.write(length_enc)

        self.proc.stdin.write(params_enc)
        self.proc.stdin.flush()

    def exec_command(self, method, *args, **kwargs):
        # protocol: len of package [8bytes], packed (pickled dict) [len]

        params = {
            'method': method,
            'args': args,
            'kwargs': kwargs
        }
        self.send_command(params)

        length_enc = self.proc.stdout.read(8)
        length = struct.unpack('L', length_enc)[0]

        ret_enc = self.proc.stdout.read(length)
        ret = pickle.loads(ret_enc)

        if ret is not None and 'protocol_error' in ret:
            raise RuntimeError(ret)
        return ret


# ================= thread to keep wrapper opened ====================

process_thread = None
queue_in = queue.Queue()
queue_out = queue.Queue()
queue_lock = Lock()


def process_keeper():
    wrapper = MLHandlerWrapper()

    while True:
        method_name, args, kwargs = queue_in.get()
        method = wrapper.__getattr__(method_name)
        try:
            ret = method(*args, **kwargs)

        except RuntimeError as e:
            ret = e
        except Exception as e:
            # TODO to something else?
            ret = e
        queue_in.task_done()
        queue_out.put(ret)


class MLHandlerPersistWrapper:
    '''
        keeps child process opened
    '''

    def __init__(self):
        global process_thread
        if process_thread is None:
            process_thread = threading.Thread(target=process_keeper, daemon=True)
            process_thread.start()

    def __getattr__(self, method_name):
        global queue_in, queue_out, queue_lock

        # is call of the method

        def call(*args, **kwargs):
            # send one task at the time
            queue_lock.acquire()

            queue_in.put([method_name, args, kwargs])
            result = queue_out.get()

            queue_lock.release()

            if isinstance(result, Exception):
                raise result

            return result

        return call

    def close(self):
        # not close child pricess
        pass


# run child process
if __name__ == '__main__':
    MLHandlerProcess().mainloop()



import bson


def is_true(val):
    return bool(val) is True


def is_false(val):
    return bool(val) is False


def int_to_objectid(n):
    s = str(n)
    s = '0' * (24 - len(s)) + s
    return bson.ObjectId(s)


def objectid_to_int(obj):
    return int(str(obj))



"""
This module defines the wrapper for ML engines which abstracts away a lot of complexity.

In particular, three big components are included:

    - `BaseMLEngineExec` class: this class wraps any object that inherits from `BaseMLEngine` and exposes some endpoints
      normally associated with a DB handler (e.g. `native_query`, `get_tables`), as well as other ML-specific behaviors,
      like `learn()` or `predict()`. Note that while these still have to be implemented at the engine level, the burden
      on that class is lesser given that it only needs to return a pandas DataFrame. It's this class that will take said
      output and format it into the HandlerResponse instance that MindsDB core expects.

    - `learn_process` method: handles async dispatch of the `learn` method in an engine, as well as registering all
      models inside of the internal MindsDB registry.

    - `predict_process` method: handles async dispatch of the `predict` method in an engine.

"""  # noqa

import time
import threading
import datetime as dt
from typing import Optional, Callable
from concurrent.futures import ProcessPoolExecutor, Future

import pandas as pd
from sqlalchemy import func, null
from sqlalchemy.sql.functions import coalesce

from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.utilities.config import Config
import mindsdb.interfaces.storage.db as db
from mindsdb.integrations.libs.response import (
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.__about__ import __version__ as mindsdb_version
from mindsdb.utilities.hooks import after_predict as after_predict_hook
from mindsdb.interfaces.model.model_controller import ModelController
from mindsdb.interfaces.model.functions import (
    get_model_record
)
from mindsdb.integrations.libs.const import PREDICTOR_STATUS
from mindsdb.interfaces.database.database import DatabaseController
from mindsdb.interfaces.storage.model_fs import ModelStorage, HandlerStorage
from mindsdb.utilities.context import context as ctx
from mindsdb.interfaces.model.functions import get_model_records
from mindsdb.integrations.handlers_client.ml_client_factory import MLClientFactory
from mindsdb.integrations.libs.learn_process import learn_process
from mindsdb.utilities.functions import mark_process
import mindsdb.utilities.profiler as profiler

from .ml_handler_proc import MLHandlerWrapper, MLHandlerPersistWrapper

import torch.multiprocessing as mp
mp_ctx = mp.get_context('spawn')


class MLEngineException(Exception):
    pass


def init_ml_handler(module_path):
    import importlib  # noqa

    from mindsdb.integrations.libs.learn_process import learn_process  # noqa

    importlib.import_module(module_path)


def dummy_task():
    return None


class WarmProcess:
    """ Class-wrapper for a process that persist for a long time. The process
        may be initialized with any handler requirements. Current implimentation
        is based on ProcessPoolExecutor just because of multiprocessing.pool
        produce daemon processes, which can not be used for learning. That
        bahaviour may be changed only using inheritance.
    """
    def __init__(self, initializer: Optional[Callable] = None, initargs: tuple = ()):
        """ create and init new process

            Args:
                initializer (Callable): the same as ProcessPoolExecutor initializer
                initargs (tuple): the same as ProcessPoolExecutor initargs
        """
        self.pool = ProcessPoolExecutor(1, initializer=initializer, initargs=initargs)
        # region bacause of ProcessPoolExecutor does not start new process
        # untill it get a task, we need manually run dummy task to force init.
        self.task = self.pool.submit(dummy_task)
        self._init_done = False
        self.task.add_done_callback(self._init_done_callback)
        # endregion

    def __del__(self):
        self.pool.shutdown(wait=False)

    def _init_done_callback(self, _task):
        """ callback for initial task
        """
        self._init_done = True

    def ready(self) -> bool:
        """ check is process ready to get a task or not

            Returns:
                bool
        """
        if self._init_done is False:
            self.task.result()
            self._init_done = True
        if self.task is None or self.task.done():
            return True
        return False

    def apply_async(self, func: Callable, *args: tuple, **kwargs: dict) -> Future:
        """ Run new task

            Args:
                func (Callable): function to run
                args (tuple): args to be passed to function
                kwargs (dict): kwargs to be passed to function

            Returns:
                Future
        """
        if not self.ready():
            raise Exception('Process task is not ready')
        self.task = self.pool.submit(
            func, *args, **kwargs
        )
        return self.task


class ProcessCache:
    """ simple cache for WarmProcess-es
    """
    def __init__(self, ttl: int = 120):
        """ Args:
            ttl (int) time to live for unused process
        """
        self.cache = {}
        self._init = False
        self._lock = threading.Lock()
        self._ttl = ttl
        self._keep_alive = {}
        self._stop_event = threading.Event()
        self.cleaner_thread = None
        self._start_clean()

    def __del__(self):
        self._stop_clean()

    def _start_clean(self) -> None:
        """ start worker that close connections after ttl expired
        """
        if (
            isinstance(self.cleaner_thread, threading.Thread)
            and self.cleaner_thread.is_alive()
        ):
            return
        self._stop_event.clear()
        self.cleaner_thread = threading.Thread(target=self._clean)
        self.cleaner_thread.daemon = True
        self.cleaner_thread.start()

    def _stop_clean(self) -> None:
        """ stop clean worker
        """
        self._stop_event.set()

    def init(self, preload_handlers: dict):
        """ run processes for specified handlers

            Args:
                preload_handlers (dict): {handler_class: count_of_processes}
        """
        with self._lock:
            if self._init is False:
                self._init = True
                for handler in preload_handlers:
                    self._keep_alive[handler.__name__] = preload_handlers[handler]
                    self.cache[handler.__name__] = {
                        'last_usage_at': time.time(),
                        'processes': [
                            WarmProcess(init_ml_handler, (handler.__module__,))
                            for _x in range(preload_handlers[handler])
                        ]
                    }

    def apply_async(self, handler: object, func: Callable, *args, **kwargs) -> Future:
        """ run new task. If possible - do it in existing process, if not - start new one.

            Args:
                handler (object): handler class
                func (Callable): function to run
                args (tuple): args to be passed to function
                kwargs (dict): kwargs to be passed to function

            Returns:
                Future
        """
        with self._lock:
            handler_name = handler.__name__
            if handler_name not in self.cache:
                warm_process = WarmProcess(init_ml_handler, (handler.__module__,))
                self.cache[handler_name] = {
                    'last_usage_at': None,
                    'processes': [warm_process]
                }
            else:
                for warm_process in self.cache[handler_name]['processes']:
                    if warm_process.ready():
                        break
                else:
                    warm_process = WarmProcess(init_ml_handler, (handler.__module__,))
                    self.cache[handler_name]['processes'].append(warm_process)
            task = warm_process.apply_async(func, *args, **kwargs)
            self.cache[handler_name]['last_usage_at'] = time.time()
        return task

    def _clean(self) -> None:
        """ worker that stop unused processes
        """
        while self._stop_event.wait(timeout=10) is False:
            with self._lock:
                for handler_name in self.cache.keys():
                    last_usage_at = self.cache[handler_name]['last_usage_at']
                    processes = self.cache[handler_name]['processes']
                    if (
                        (
                            handler_name not in self._keep_alive
                            or self._keep_alive[handler_name] < len(processes)
                        )
                        and last_usage_at is not None
                        and (time.time() - last_usage_at) > self._ttl
                    ):
                        for i, process in enumerate(processes):
                            if process.ready():
                                processes.pop(i)
                                del process
                                break


process_cache = ProcessCache()


class BaseMLEngineExec:

    def __init__(self, name, **kwargs):
        """
        ML handler interface converter
        """  # noqa
        # TODO move this class to model controller

        self.name = name
        self.config = Config()
        self.handler_controller = kwargs.get('handler_controller')
        self.company_id = kwargs.get('company_id')
        self.fs_store = kwargs.get('file_storage')
        self.storage_factory = kwargs.get('storage_factory')
        self.integration_id = kwargs.get('integration_id')
        self.execution_method = kwargs.get('execution_method')
        self.engine = kwargs.get("integration_engine")

        self.model_controller = ModelController()
        self.database_controller = DatabaseController()

        self.parser = parse_sql
        self.dialect = 'mindsdb'

        self.is_connected = True

        self.handler_class = MLClientFactory(handler_class=kwargs['handler_class'], engine=self.engine)

    def _get_ml_handler(self, predictor_id=None):
        # returns instance or wrapper over it

        integration_id = self.integration_id

        class_path = [self.handler_class.__module__, self.handler_class.__name__]

        if self.execution_method == 'subprocess':
            handler = MLHandlerWrapper()

            handler.init_handler(class_path, integration_id, predictor_id, ctx.dump())
            return handler

        elif self.execution_method == 'subprocess_keep':
            handler = MLHandlerPersistWrapper()

            handler.init_handler(class_path, integration_id, predictor_id, ctx.dump())
            return handler

        elif self.execution_method == 'remote':
            raise NotImplementedError()

        else:
            handlerStorage = HandlerStorage(integration_id)
            modelStorage = ModelStorage(predictor_id)

            ml_handler = self.handler_class(
                engine_storage=handlerStorage,
                model_storage=modelStorage,
            )
            return ml_handler

    def get_tables(self) -> Response:
        """ Returns all models currently registered that belong to the ML engine."""
        all_models = self.model_controller.get_models(integration_id=self.integration_id)
        all_models_names = [[x['name']] for x in all_models]
        response = Response(
            RESPONSE_TYPE.TABLE,
            pd.DataFrame(
                all_models_names,
                columns=['table_name']
            )
        )
        return response

    def get_columns(self, table_name: str) -> Response:
        """ Retrieves standard info about a model, e.g. data types. """  # noqa
        predictor_record = get_model_record(name=table_name, ml_handler_name=self.name)
        if predictor_record is None:
            return Response(
                RESPONSE_TYPE.ERROR,
                error_message=f"Error: model '{table_name}' does not exist!"
            )

        data = []
        if predictor_record.dtype_dict is not None:
            for key, value in predictor_record.dtype_dict.items():
                data.append((key, value))
        result = Response(
            RESPONSE_TYPE.TABLE,
            pd.DataFrame(
                data,
                columns=['COLUMN_NAME', 'DATA_TYPE']
            )
        )
        return result

    def native_query(self, query: str) -> Response:
        """ Intakes a raw SQL query and returns the answer given by the ML engine. """
        query_ast = self.parser(query, dialect=self.dialect)
        return self.query(query_ast)

    def query_(self, query: ASTNode) -> Response:
        raise Exception('Should not be used')

    @profiler.profile()
    def learn(
        self, model_name, project_name,
        data_integration_ref=None,
        fetch_data_query=None,
        problem_definition=None,
        join_learn_process=False,
        label=None,
        is_retrain=False,
        set_active=True,
    ):
        # TODO move to model_controller
        """ Trains a model given some data-gathering SQL statement. """

        target = problem_definition['target']

        project = self.database_controller.get_project(name=project_name)

        # handler-side validation
        # self.handler_class is a instance of MLClientFactory
        # so need to check self.handler_class.handler_class attribute
        # which is a class of a real MLHandler
        if hasattr(self.handler_class.handler_class, 'create_validation'):
            self.handler_class.handler_class.create_validation(target, args=problem_definition)

        predictor_record = db.Predictor(
            company_id=ctx.company_id,
            name=model_name,
            integration_id=self.integration_id,
            data_integration_ref=data_integration_ref,
            fetch_data_query=fetch_data_query,
            mindsdb_version=mindsdb_version,
            to_predict=target,
            learn_args=problem_definition,
            data={'name': model_name},
            project_id=project.id,
            training_data_columns_count=None,
            training_data_rows_count=None,
            training_start_at=dt.datetime.now(),
            status=PREDICTOR_STATUS.GENERATING,
            label=label,
            version=(
                db.session.query(
                    coalesce(func.max(db.Predictor.version), 1) + (1 if is_retrain else 0)
                ).filter_by(
                    company_id=ctx.company_id,
                    name=model_name,
                    project_id=project.id,
                    deleted_at=null()
                ).scalar_subquery()),
            active=(not is_retrain),  # if create then active
        )

        db.serializable_insert(predictor_record)

        class_path = [self.handler_class.__module__, self.handler_class.__name__]

        task = process_cache.apply_async(
            self.handler_class,
            learn_process,
            class_path,
            self.engine,
            ctx.dump(),
            self.integration_id,
            predictor_record.id,
            problem_definition,
            set_active,
            data_integration_ref=data_integration_ref,
            fetch_data_query=fetch_data_query,
            project_name=project_name
        )

        if join_learn_process is True:
            task.result()
            predictor_record = db.Predictor.query.get(predictor_record.id)
            db.session.refresh(predictor_record)

        return predictor_record

    @profiler.profile()
    @mark_process(name='predict')
    def predict(self, model_name: str, data: list, pred_format: str = 'dict',
                project_name: str = None, version=None, params: dict = None):
        """ Generates predictions with some model and input data. """
        if isinstance(data, dict):
            data = [data]
        df = pd.DataFrame(data)
        kwargs = {
            'name': model_name,
            'ml_handler_name': self.name,
            'project_name': project_name
        }
        if version is None:
            kwargs['active'] = True
        else:
            kwargs['active'] = None
            kwargs['version'] = version
        predictor_record = get_model_record(**kwargs)
        if predictor_record is None:
            if version is not None:
                model_name = f'{model_name}.{version}'
            raise Exception(f"Error: model '{model_name}' does not exists!")
        if predictor_record.status != PREDICTOR_STATUS.COMPLETE:
            raise Exception("Error: model creation not completed")

        ml_handler = self._get_ml_handler(predictor_record.id)

        args = {
            'pred_format': pred_format,
            'predict_params': {} if params is None else params
        }
        # FIXME
        # if self.handler_class.__name__ == 'LightwoodHandler':
        if self.handler_class.__name__ == 'LightwoodHandler':
            args['code'] = predictor_record.code
            args['target'] = predictor_record.to_predict[0]
            args['dtype_dict'] = predictor_record.dtype_dict
            args['learn_args'] = predictor_record.learn_args

        if self.handler_class.__name__ in ('LangChainHandler',):
            from mindsdb.api.mysql.mysql_proxy.controllers import SessionController
            from mindsdb.api.mysql.mysql_proxy.executor.executor_commands import ExecuteCommands

            sql_session = SessionController()
            sql_session.database = 'mindsdb'

            command_executor = ExecuteCommands(sql_session, executor=None)

            args['executor'] = command_executor

        try:
            predictions = ml_handler.predict(df, args)
        except Exception as e:
            msg = str(e).strip()
            if msg == '':
                msg = e.__class__.__name__
            msg = f'[{self.name}/{model_name}]: {msg}'
            raise MLEngineException(msg) from e

        ml_handler.close()

        # mdb indexes
        if '__mindsdb_row_id' not in predictions.columns and '__mindsdb_row_id' in df.columns:
            predictions['__mindsdb_row_id'] = df['__mindsdb_row_id']

        after_predict_hook(
            company_id=self.company_id,
            predictor_id=predictor_record.id,
            rows_in_count=df.shape[0],
            columns_in_count=df.shape[1],
            rows_out_count=len(predictions)
        )
        return predictions

    @profiler.profile()
    def update(
            self, model_name, project_name,
            base_model_version: int,
            data_integration_ref=None,
            fetch_data_query=None,
            join_learn_process=False,
            label=None,
            set_active=True,
            args: Optional[dict] = None
    ):
        # generate new record from latest version as starting point
        project = self.database_controller.get_project(name=project_name)

        search_args = {
            'active': None,
            'name': model_name,
            'status': PREDICTOR_STATUS.COMPLETE
        }
        if base_model_version is not None:
            search_args['version'] = base_model_version
        else:
            search_args['active'] = True
        predictor_records = get_model_records(**search_args)
        if len(predictor_records) == 0:
            raise Exception("Can't find suitable base model")

        predictor_records.sort(key=lambda x: x.training_stop_at, reverse=True)
        predictor_records = [x for x in predictor_records if x.training_stop_at is not None]
        base_predictor_record = predictor_records[0]

        learn_args = base_predictor_record.learn_args
        learn_args['using'] = args if not learn_args.get('using', False) else {**learn_args['using'], **args}

        predictor_record = db.Predictor(
            company_id=ctx.company_id,
            name=model_name,
            integration_id=self.integration_id,
            data_integration_ref=data_integration_ref,
            fetch_data_query=fetch_data_query,
            mindsdb_version=mindsdb_version,
            to_predict=base_predictor_record.to_predict,
            learn_args=learn_args,
            data={'name': model_name},
            project_id=project.id,
            training_data_columns_count=None,
            training_data_rows_count=None,
            training_start_at=dt.datetime.now(),
            status=PREDICTOR_STATUS.GENERATING,
            label=label,
            version=(
                db.session.query(
                    coalesce(func.max(db.Predictor.version), 1) + 1
                ).filter_by(
                    company_id=ctx.company_id,
                    name=model_name,
                    project_id=project.id,
                    deleted_at=null()
                ).scalar_subquery()
            ),
            active=False
        )
        with profiler.Context('finetune-update-record-insert'):
            db.serializable_insert(predictor_record)

        class_path = [self.handler_class.__module__, self.handler_class.__name__]

        with profiler.Context('finetune-update'):
            task = process_cache.apply_async(
                self.handler_class,
                learn_process,
                class_path,
                self.engine,
                ctx.dump(),
                self.integration_id,
                predictor_record.id,
                predictor_record.learn_args,
                set_active,
                base_predictor_record.id,
                data_integration_ref=data_integration_ref,
                fetch_data_query=fetch_data_query,
                project_name=project_name
            )

            if join_learn_process is True:
                task.result()
                predictor_record = db.Predictor.query.get(predictor_record.id)
                db.session.refresh(predictor_record)

        return predictor_record



from collections import OrderedDict
from typing import Optional
import pandas as pd
import pymysql as matone
from pymysql.cursors import DictCursor as dict
from mindsdb_sql import parse_sql
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb_sql.parser.ast.base import ASTNode

from mindsdb.utilities import log
from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE


class MatrixOneHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the MatrixOne statements.
    """

    name = 'matrixone'

    def __init__(self, name,connection_data: Optional[dict], **kwargs):
        super().__init__(name)
        self.mysql_url = None
        self.parser = parse_sql
        self.dialect = 'mysql'
        self.connection_data = connection_data
        self.database = self.connection_data.get('database')

        self.connection = None
        self.is_connected = False

    def connect(self):
        if self.is_connected is True:
            return self.connection

        config = {
            'host': self.connection_data.get('host'),
            'port': self.connection_data.get('port'),
            'user': self.connection_data.get('user'),
            'password': self.connection_data.get('password'),
            'database': self.connection_data.get('database')
        }

        ssl = self.connection_data.get('ssl')
        if ssl is True:
            ssl_ca = self.connection_data.get('ssl_ca')
            ssl_cert = self.connection_data.get('ssl_cert')
            ssl_key = self.connection_data.get('ssl_key')
            config['client_flags'] = [matone.constants.ClientFlag.SSL]
            if ssl_ca is not None:
                config["ssl_ca"] = ssl_ca
            if ssl_cert is not None:
                config["ssl_cert"] = ssl_cert
            if ssl_key is not None:
                config["ssl_key"] = ssl_key

        connection = matone.connect(**config)
        self.is_connected = True
        self.connection = connection
        return self.connection

    def disconnect(self):
        if self.is_connected is False:
            return
        self.connection.close()
        self.is_connected = False
        return

    def check_connection(self) -> StatusResponse:
        """
        Check the connection of the MatrixOne database
        :return: success status and error message if error occurs
        """

        result = StatusResponse(False)
        need_to_close = self.is_connected is False

        try:
            connection = self.connect()
            result.success = connection.open
        except Exception as e:
            log.logger.error(f'Error connecting to MatrixOne {self.connection_data["database"]}, {e}!')
            result.error_message = str(e)

        if result.success is True and need_to_close:
            self.disconnect()
        if result.success is False and self.is_connected is True:
            self.is_connected = False

        return result

    def native_query(self, query: str) -> Response:
        """
        Receive SQL query and runs it
        :param query: The SQL query to run in MatrixOne
        :return: returns the records from the current recordset
        """

        need_to_close = self.is_connected is False

        connection = self.connect()
        with connection.cursor(cursor=dict) as cur:
            try:
                cur.execute(query)
                if cur._rows:
                    result = cur.fetchall()
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        pd.DataFrame(
                            result,
                            # columns=[x[0] for x in cur.description]
                        )
                    )
                else:
                    response = Response(RESPONSE_TYPE.OK)
                connection.commit()
            except Exception as e:
                log.logger.error(f'Error running query: {query} on {self.connection_data["database"]}!')
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_message=str(e)
                )
                connection.rollback()

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """
        Retrieve the data from the SQL statement.
        """
        renderer = SqlalchemyRender('mysql')
        query_str = renderer.get_string(query, with_failback=True)
        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """
        Get a list with all of the tabels in MatrixOne
        """
        q = "SHOW TABLES;"
        result = self.native_query(q)
        df = result.data_frame
        result.data_frame = df.rename(columns={df.columns[0]: 'table_name'})
        return result

    def get_columns(self, table_name) -> Response:
        """
        Show details about the table
        """
        q = f"SHOW COLUMNS FROM {table_name};"
        result = self.native_query(q)
        df=result.data_frame
        result.data_frame=df.rename(columns={
            df.columns[0]: 'COLUMN_NAME',
            df.columns[1]: 'DATA TYPE'
        })


        return result


connection_args = OrderedDict(
    user={
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the MatrixOne server.'
    },
    password={
        'type': ARG_TYPE.STR,
        'description': 'The password to authenticate the user with the MatrixOne server.'
    },
    database={
        'type': ARG_TYPE.STR,
        'description': 'The database name to use when connecting with the MatrixOne server.'
    },
    host={
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the MatrixOne server. '
    },
    port={
        'type': ARG_TYPE.INT,
        'description': 'The TCP/IP port of the MatrixOne server. Must be an integer.'
    },
    ssl={
        'type': ARG_TYPE.BOOL,
        'description': 'Set it to False to disable ssl.'
    },
    ssl_ca={
        'type': ARG_TYPE.PATH,
        'description': 'Path or URL of the Certificate Authority (CA) certificate file'
    },
    ssl_cert={
        'type': ARG_TYPE.PATH,
        'description': 'Path name or URL of the server public key certificate file'
    },
    ssl_key={
        'type': ARG_TYPE.PATH,
        'description': 'The path name or URL of the server private key file'
    }
)

connection_args_example = OrderedDict(
    host='127.0.0.1',
    port=6001,
    user='dump',
    password='111',
    database='mo_catalog'
)



"""training_time

Revision ID: 6e834843e7e9
Revises: b5b53e0ea7f8
Create Date: 2022-07-22 13:32:34.796604

"""
from alembic import op
import sqlalchemy as sa
import mindsdb.interfaces.storage.db    # noqa


# revision identifiers, used by Alembic.
revision = '6e834843e7e9'
down_revision = 'b5b53e0ea7f8'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.add_column(sa.Column('training_start_at', sa.DateTime(), nullable=True))
        batch_op.add_column(sa.Column('training_stop_at', sa.DateTime(), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('predictor', schema=None) as batch_op:
        batch_op.drop_column('training_stop_at')
        batch_op.drop_column('training_start_at')
    # ### end Alembic commands ###



import unittest
from mindsdb.integrations.handlers.vertica_handler.vertica_handler import VerticaHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class VerticaHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.kwargs = {
            "connection_data": {
                "host":'127.0.0.1',
                "port":5433,
                "user":'dbadmin',
                "password":'',
                "database":'VMart',
                "schema_name":'public'
            }
    
        }
        cls.handler = VerticaHandler('test_vertica_handler', cls.kwargs)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_connect(self):
        assert self.handler.connect()
    
    def test_2_create_table(self):
        query = "CREATE Table TEST(id Number(1),Name Varchar(33))"
        result = self.handler.query(query)
        assert result.type is not RESPONSE_TYPE.ERROR 

    def test_3_insert(self):
        query = "INSERT INTO TEST (1,'lOVe yOU)"
        result = self.handler.query(query)
        assert result.type is not RESPONSE_TYPE.ERROR 

    def test_4_native_query_select(self):
        query = "SELECT * FROM TEST;"
        result = self.handler.query(query)
        assert result.type is RESPONSE_TYPE.TABLE 

    def test_5_get_tables(self):
        tables = self.handler.get_tables()
        assert tables.type is not RESPONSE_TYPE.TABLE

    def test_6_get_columns(self):
        columns = self.handler.get_columns('TEMP')
        assert columns.type is not RESPONSE_TYPE.ERROR


if __name__ == '__main__':
    unittest.main()



import unittest
from mindsdb.integrations.handlers.huggingface_api_handler.huggingface_api_handler import HuggingFaceInferenceAPIHandler
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE


class HuggingFaceAPIHandlerTest(unittest.TestCase):
    pass


if __name__ == '__main__':
    unittest.main()



import copy

import duckdb
import numpy as np

from mindsdb_sql import parse_sql
from mindsdb_sql.render.sqlalchemy_render import SqlalchemyRender
from mindsdb_sql.planner.utils import query_traversal
from mindsdb_sql.parser.ast import (
    Select, Identifier,
    Function, Constant
)

from mindsdb.utilities import log
from mindsdb.utilities.json_encoder import CustomJSONEncoder


def query_df(df, query, session=None):
    """ Perform simple query ('select' from one table, without subqueries and joins) on DataFrame.

        Args:
            df (pandas.DataFrame): data
            query (mindsdb_sql.parser.ast.Select | str): select query

        Returns:
            pandas.DataFrame
    """

    if isinstance(query, str):
        query_ast = parse_sql(query, dialect='mysql')
    else:
        query_ast = copy.deepcopy(query)

    if isinstance(query_ast, Select) is False \
       or isinstance(query_ast.from_table, Identifier) is False:
        raise Exception(
            "Only 'SELECT from TABLE' statements supported for internal query"
        )

    table_name = query_ast.from_table.parts[0]
    query_ast.from_table.parts = ['df_table']

    json_columns = set()

    def adapt_query(node, is_table, **kwargs):
        if is_table:
            return
        if isinstance(node, Identifier):
            if len(node.parts) > 1:
                node.parts = [node.parts[-1]]
                return node
        if isinstance(node, Function):
            fnc_name = node.op.lower()
            if fnc_name == 'database' and len(node.args) == 0:
                if session is not None:
                    cur_db = session.database
                else:
                    cur_db = None
                return Constant(cur_db)
            if fnc_name == 'truncate':
                # replace mysql 'truncate' function to duckdb 'round'
                node.op = 'round'
                if len(node.args) == 1:
                    node.args.append(0)
            if fnc_name == 'json_extract':
                json_columns.add(node.args[0].parts[-1])

    query_traversal(query_ast, adapt_query)

    # convert json columns
    encoder = CustomJSONEncoder()

    def _convert(v):
        if isinstance(v, dict) or isinstance(v, list):
            try:
                return encoder.encode(v)
            except Exception:
                pass
        return v
    for column in json_columns:
        df[column] = df[column].apply(_convert)

    render = SqlalchemyRender('postgres')
    try:
        query_str = render.get_string(query_ast, with_failback=False)
    except Exception as e:
        log.logger.error(
            f"Exception during query casting to 'postgres' dialect. Query: {str(query)}. Error: {e}"
        )
        query_str = render.get_string(query_ast, with_failback=True)

    # workaround to prevent duckdb.TypeMismatchException
    if len(df) > 0 and table_name.lower() in ('models', 'predictors', 'models_versions'):
        if 'TRAINING_OPTIONS' in df.columns:
            df = df.astype({'TRAINING_OPTIONS': 'string'})

    con = duckdb.connect(database=':memory:')
    con.register('df_table', df)
    result_df = con.execute(query_str).fetchdf()
    result_df = result_df.replace({np.nan: None})
    description = con.description
    con.unregister('df_table')
    con.close()

    new_column_names = {}
    real_column_names = [x[0] for x in description]
    for i, duck_column_name in enumerate(result_df.columns):
        new_column_names[duck_column_name] = real_column_names[i]
    result_df = result_df.rename(
        new_column_names,
        axis='columns'
    )
    return result_df



import unittest
from mindsdb.integrations.handlers.bigquery_handler.bigquery_handler import BigQueryHandler
from mindsdb.integrations.libs.response import RESPONSE_TYPE


class BigQueryHandlerTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        connection_data = {            
            "project_id": "tough-future-332513",
            "service_account_keys": "/home/bigq/tough-future-332513-6699d9b39601.json"
        }
        cls.handler = BigQueryHandler('test_bigquery_handler', connection_data)

    def test_0_check_connection(self):
        assert self.handler.check_connection()

    def test_1_list_tables(self):        
        datasets = self.handler.get_tables('Rentals')
        assert datasets.type is not RESPONSE_TYPE.ERROR

    def test_2_list_table_columns(self):        
        datasets = self.handler.get_columns('Rentals', 'home_rentals')
        assert datasets.type is not RESPONSE_TYPE.ERROR


from collections import OrderedDict
from typing import Any, Optional

import jaydebeapi as jdbcconnector
from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast.base import ASTNode
import pandas as pd

from mindsdb.integrations.libs.base import DatabaseHandler
from mindsdb.integrations.libs.const import HANDLER_CONNECTION_ARG_TYPE as ARG_TYPE
from mindsdb.integrations.libs.response import (
    HandlerStatusResponse as StatusResponse,
    HandlerResponse as Response,
    RESPONSE_TYPE
)
from mindsdb.utilities import log




class AltibaseHandler(DatabaseHandler):
    """
    This handler handles connection and execution of the Altibase statements.
    """

    name = 'altibase'

    def __init__(self, name: str, connection_data: Optional[dict]):
        """ constructor
        Args:
            name (str): name of particular handler instance
            connection_data (dict): parameters for connecting to the database
        """
        super().__init__(name)
        
        self.parser = parse_sql

        self.connection_args = connection_data
        self.database = self.connection_args.get('database')
        self.host = self.connection_args.get('host')
        self.port = self.connection_args.get('port')
        
        self.connection = None
        self.is_connected = False

    def connect(self):
        """ Set up any connections required by the handler
        Should return output of check_connection() method after attempting connection.
        Should switch self.is_connected.
        Returns:
            connection
        """
        if self.is_connected is True:
            return self.connection
                
        user = self.connection_args.get('user')
        password = self.connection_args.get('password')
        jar_location = self.connection_args.get('jdbcJarLocation')

        jdbc_class = self.connection_args.get('jdbcClass', 'Altibase.jdbc.driver.AltibaseDriver')
        jdbc_url = f"jdbc:Altibase://{self.host}:{self.port}/{self.database}"

        try:
            if user and password and jar_location: 
                connection = jdbcconnector.connect(jclassname=jdbc_class, url=jdbc_url, driver_args=[user, password], jars=jar_location.split(","))
            elif user and password: 
                connection = jdbcconnector.connect(jclassname=jdbc_class, url=jdbc_url, driver_args=[user, password])
            elif jar_location: 
                connection = jdbcconnector.connect(jclassname=jdbc_class, url=jdbc_url, jars=jar_location.split(","))
            else:
                connection = jdbcconnector.connect(jclassname=jdbc_class, url=jdbc_url)

            self.connection = connection
            self.is_connected = True
        except Exception as e:
            log.logger.error(f"Error while connecting to {self.database}, {e}")
        
        return self.connection

    def disconnect(self):
        """ Close any existing connections
        Should switch self.is_connected.
        """
        if self.is_connected is True:
            try:
                self.connection.close()
                self.is_connected = False
            except Exception as e:
                log.logger.error(f"Error while disconnecting to {self.database}, {e}")
                return False
        return True 

    def check_connection(self) -> StatusResponse:
        """ Check connection to the handler
        Returns:
            HandlerStatusResponse
        """
        responseCode = StatusResponse(success=False)
        need_to_close = self.is_connected is False

        try:
            self.connect()
            responseCode.success = True
        except Exception as e:
            log.logger.error(f'Error connecting to database {self.database}, {e}!')
            responseCode.error_message = str(e)
        finally:
            if responseCode.success and need_to_close:
                self.disconnect()
            if not responseCode.success and self.is_connected:
                self.is_connected = False

        return responseCode

    def native_query(self, query: str) -> Response:
        """Receive raw query and act upon it somehow.
        Args:
            query (str): query in native format
        Returns:
            HandlerResponse
        """
        need_to_close = self.is_connected is False
        connection = self.connect()
        with connection.cursor() as cur:
            try:
                cur.execute(query)
                if cur.description:
                    result = cur.fetchall() 
                    response = Response(
                        RESPONSE_TYPE.TABLE,
                        data_frame=pd.DataFrame(
                            result,
                            columns=[x[0] for x in cur.description]
                        )
                    )
                else:
                    response = Response(RESPONSE_TYPE.OK)
                connection.commit()
            except Exception as e:
                log.logger.error(f'Error running query: {query} on {self.database}!')
                response = Response(
                    RESPONSE_TYPE.ERROR,
                    error_message=str(e)
                )
                connection.rollback()

        if need_to_close is True:
            self.disconnect()

        return response

    def query(self, query: ASTNode) -> Response:
        """Receive query as AST (abstract syntax tree) and act upon it somehow.
        Args:
            query (ASTNode): sql query represented as AST. May be any kind of query: SELECT, INSERT, DELETE, etc
        Returns:
            HandlerResponse
        """
        if isinstance(query, ASTNode):
            query_str = query.to_string()
        else:
            query_str = str(query)

        return self.native_query(query_str)

    def get_tables(self) -> Response:
        """ Return list of entities
        Return list of entities that will be accesible as tables.
        Returns:
            HandlerResponse
        """
        query = '''
            SELECT 
                TABLE_NAME, 
                TABLE_ID, 
                TABLE_TYPE 
            FROM 
                system_.sys_tables_ 
            WHERE 
                user_id = USER_ID();
            '''
    
        return self.native_query(query)

    def get_columns(self, table_name: str) -> Response:
        """ Returns a list of entity columns
        Args:
            table_name (str): name of one of tables returned by self.get_tables()
        Returns:
            HandlerResponse
        """
        query = f"""
            SELECT 
                COLUMN_NAME, 
                DATA_TYPE 
            FROM 
                system_.sys_columns_ ct 
            inner join 
                system_.sys_tables_ tt 
                on ct.table_id=tt.table_id 
            where 
                tt.table_name = '{table_name.capitalize()}';
            """
        
        return self.native_query(query)
    
connection_args = OrderedDict(
    host = {
        'type': ARG_TYPE.STR,
        'description': 'The host name or IP address of the Altibase server. NOTE: use \'127.0.0.1\' instead of \'localhost\' to connect to local server.'
    },
    port = {
        'type': ARG_TYPE.INT,
        'description': 'The TCP/IP port of the Altibase server. Must be an integer.'
    },
    user = {
        'type': ARG_TYPE.STR,
        'description': 'The user name used to authenticate with the Altibase server.'
    },
    password = {
        'type': ARG_TYPE.STR,
        'description': 'The password to authenticate the user with the Altibase server.'
    },
    database = {
        'type': ARG_TYPE.STR,
        'description': 'The database name to use when connecting with the Altibase server.'
    },
    jdbcClass = {
        'type': ARG_TYPE.STR,
        'description': 'The driver class of the Altibase JDBC driver'
    },
    jdbcJarLocation = {
        'type': ARG_TYPE.PATH,
        'description': 'The location of the Altibase JDBC driver jar file'
    },
)

connection_args_example = OrderedDict(
    host = '127.0.0.1',
    port = 20300,
    user = 'sys',
    password = 'manager',
    database = 'mydb',
    jdbcClass='Altibase.jdbc.driver.AltibaseDriver',
    jdbcJarLocation='/data/altibase_home/lib/Altibase.jar'
)



import time

import pandas as pd
from flask import request
from flask_restx import Resource
from pandas.core.frame import DataFrame
from dataprep_ml.insights import analyze_dataset
from mindsdb_sql import parse_sql
from mindsdb_sql.parser.ast import Constant, Identifier
from mindsdb_sql.planner.utils import query_traversal

from mindsdb.api.http.utils import http_error
from mindsdb.api.http.namespaces.configs.analysis import ns_conf
from mindsdb.api.mysql.mysql_proxy.classes.fake_mysql_proxy import FakeMysqlProxy
from mindsdb.api.mysql.mysql_proxy.libs.constants.response_type import RESPONSE_TYPE as SQL_RESPONSE_TYPE


def analyze_df(df: DataFrame) -> dict:
    if len(df) == 0:
        return {}

    cols = pd.Series(df.columns)

    # https://stackoverflow.com/questions/24685012/pandas-dataframe-renaming-multiple-identically-named-columns
    for dup in cols[cols.duplicated()].unique():
        cols[cols[cols == dup].index.values.tolist()] = [dup + '.' + str(i) if i != 0 else dup for i in
                                                         range(sum(cols == dup))]

    # rename the columns with the cols list.
    df.columns = cols

    analysis = analyze_dataset(df)
    return analysis.to_dict()


@ns_conf.route('/query')
class QueryAnalysis(Resource):
    @ns_conf.doc('post_query_to_analyze')
    def post(self):
        data = request.json
        query = data.get('query')
        context = data.get('context', {})
        limit = data.get('limit')
        if query is None or len(query) == 0:
            return http_error(400, 'Missed query', 'Need provide query to analyze')

        try:
            ast = parse_sql(query)
        except Exception as e:
            return http_error(500, 'Wrong query', str(e))

        if limit is not None:
            ast.limit = Constant(limit)
            query = str(ast)

        mysql_proxy = FakeMysqlProxy()
        mysql_proxy.set_context(context)

        try:
            result = mysql_proxy.process_query(query)
        except Exception as e:
            import traceback
            print(traceback.format_exc())
            return http_error(500, 'Error', str(e))

        if result.type == SQL_RESPONSE_TYPE.ERROR:
            return http_error(500, f'Error {result.error_code}', result.error_message)
        if result.type != SQL_RESPONSE_TYPE.TABLE:
            return http_error(500, 'Error', 'Query does not return data')

        column_names = [x['name'] for x in result.columns]
        df = DataFrame(result.data, columns=column_names)
        analysis = analyze_df(df)

        query_tables = []

        def find_tables(node, is_table, **kwargs):
            if is_table and isinstance(node, Identifier):
                query_tables.append('.'.join(node.parts))
        query_traversal(ast, find_tables)

        return {
            'analysis': analysis,
            'column_names': column_names,
            'row_count': len(result.data),
            'timestamp': time.time(),
            'tables': query_tables
        }


@ns_conf.route('/data')
class DataAnalysis(Resource):
    @ns_conf.doc('post_data_to_analyze')
    def post(self):
        payload = request.json
        column_names = payload.get('column_names')
        data = payload.get('data')

        analysis = analyze_df(DataFrame(data, columns=column_names))

        return {
            'analysis': analysis,
            'timestamp': time.time()
        }



from typing import Dict, Optional

import pandas as pd

from mindsdb.integrations.libs.base import BaseMLEngine
from mindsdb.utilities.log import get_log

from .ingest import Ingestor
from .question_answer import QuestionAnswerer
from .settings import (
    DEFAULT_EMBEDDINGS_MODEL,
    USER_DEFINED_MODEL_PARAMS,
    ModelParameters,
)

# these require no additional arguments

logger = get_log(logger_name=__name__)


class WriterHandler(BaseMLEngine):
    """
    WriterHandler is a MindsDB integration with Writer API LLMs that allows users to run question answering
    on their data by providing a question.

    The User is able to provide data that provides context for the questions, see create() method for more details.

    """

    name = "writer"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.generative = True

    @staticmethod
    def create_validation(target, args=None, **kwargs):
        if "using" not in args:
            raise Exception(
                "Writer engine requires a USING clause! Refer to its documentation for more details."
            )
        else:
            args = args["using"]

        if "prompt_template" not in args:
            raise Exception("Please provide a `prompt_template` for this engine.")

    def create(
        self,
        target: str,
        df: pd.DataFrame = pd.DataFrame(),
        args: Optional[Dict] = None,
    ):
        """
        Dispatch is running embeddings and storing in Chroma VectorDB, unless user already has embeddings persisted
        """
        args = args["using"]

        if not df.empty and args["run_embeddings"]:
            if "context_columns" not in args:
                # if no context columns provided, use all columns in df
                args["context_columns"] = df.columns.tolist()

            if "embeddings_model_name" not in args:
                logger.info(
                    f"No embeddings model provided in query, using default model: {DEFAULT_EMBEDDINGS_MODEL}"
                )

            chromadb_folder_name = args["chromadb_folder_name"]
            # create folder for chromadb to persist embeddings
            args["chromadb_storage_path"] = self.engine_storage.folder_get(
                chromadb_folder_name
            )
            ingestor = Ingestor(df=df, args=args)
            ingestor.embeddings_to_vectordb()

            # for mindsdb cloud, store data in shared file system for cloud version of mindsdb to make it be usable by all mindsdb nodes
            self.engine_storage.folder_sync(chromadb_folder_name)

        else:
            logger.info("Skipping embeddings and ingestion into Chroma VectorDB")

        self.model_storage.json_set("args", args)

    def predict(self, df: pd.DataFrame = None, args: dict = None):
        """
        Dispatch is performed depending on the underlying model type. Currently, only question answering
        is supported.
        """

        # get model parameters if defined by user - else use default values

        args = self.model_storage.json_get("args")
        args["chromadb_storage_path"] = self.engine_storage.folder_get(
            args["chromadb_folder_name"], update=False
        )

        user_defined_model_params = list(
            filter(lambda x: x in args, USER_DEFINED_MODEL_PARAMS)
        )
        args["model_params"] = {
            model_param: args[model_param] for model_param in user_defined_model_params
        }
        model_parameters = ModelParameters(**args["model_params"])

        # get question answering results

        question_answerer = QuestionAnswerer(
            args=args, model_parameters=model_parameters
        )

        # get question from sql query e.g. where question = 'What is the capital of France?'
        question_answerer.query(df["question"].tolist()[0])

        # return results

        return question_answerer.results_df



import re
import operator


def f_and(*args):
    for i in args:
        if not i:
            return False
    return True


def f_or(*args):
    for i in args:
        if i:
            return True
    return False


def f_like(s, p):
    p = '^{}$'.format(p.replace('%', r'[\s\S]*'))

    return re.match(p, s) is not None


def f_add(*args):
    # strings and numbers are supported
    # maybe it is not true sql-way

    out = args[0] + args[1]
    for i in args[2:]:
        out += i
    return out


def f_ne(a, b):
    if a is None or b is None:
        return False
    return operator.ne(a, b)


def f_eq(a, b):
    if a is None or b is None:
        return False
    return operator.eq(a, b)


operator_map = {
    '+': f_add,
    '-': operator.sub,
    '/': operator.truediv,
    '*': operator.mul,
    '%': operator.mod,
    '=': f_eq,
    '!=': f_ne,
    '>': operator.gt,
    '<': operator.lt,
    '>=': operator.ge,
    '<=': operator.le,
    'IS': operator.eq,
    'IS NOT': operator.ne,
    'LIKE': f_like,
    'NOT LIKE': lambda s, p: not f_like(s, p),
    'IN': lambda v, ll: v in ll,
    'NOT IN': lambda v, ll: v not in ll,
    'AND': f_and,
    'OR': f_or,
    '||': f_add
    # binary and, binary not, exists, missing, etc
}



import os
import sys
import time
import json
import atexit
import signal
import psutil
import asyncio
import secrets
import traceback
import threading
from textwrap import dedent
from packaging import version

from mindsdb.__about__ import __version__ as mindsdb_version
from mindsdb.api.http.start import start as start_http
from mindsdb.api.mysql.start import start as start_mysql
from mindsdb.api.mongo.start import start as start_mongo
from mindsdb.api.postgres.start import start as start_postgres
from mindsdb.interfaces.chatbot.chatbot_monitor import start as start_chatbot
from mindsdb.interfaces.jobs.scheduler import start as start_scheduler
from mindsdb.utilities.config import Config
from mindsdb.utilities.ps import is_pid_listen_port, get_child_pids
from mindsdb.utilities.functions import args_parse, get_versions_where_predictors_become_obsolete
from mindsdb.utilities import log
from mindsdb.interfaces.database.integrations import integration_controller
import mindsdb.interfaces.storage.db as db
from mindsdb.integrations.utilities.install import install_dependencies
from mindsdb.utilities.fs import create_dirs_recursive, clean_process_marks, clean_unlinked_process_marks
from mindsdb.utilities.telemetry import telemetry_file_exists, disable_telemetry
from mindsdb.utilities.context import context as ctx
from mindsdb.utilities.auth import register_oauth_client, get_aws_meta_data

import torch.multiprocessing as mp
try:
    mp.set_start_method('spawn')
except RuntimeError:
    log.logger.info('Torch multiprocessing context already set, ignoring...')


_stop_event = threading.Event()


def close_api_gracefully(apis):
    _stop_event.set()
    try:
        for api in apis.values():
            process = api['process']
            childs = get_child_pids(process.pid)
            for p in childs:
                try:
                    os.kill(p, signal.SIGTERM)
                except Exception:
                    p.kill()
            sys.stdout.flush()
            process.terminate()
            process.join()
            sys.stdout.flush()
    except KeyboardInterrupt:
        sys.exit(0)
    except psutil.NoSuchProcess:
        pass


def do_clean_process_marks():
    while _stop_event.wait(timeout=5) is False:
        clean_unlinked_process_marks()


if __name__ == '__main__':
    # ----------------  __init__.py section ------------------
    clean_process_marks()
    ctx.set_default()
    args = args_parse()

    # ---- CHECK SYSTEM ----
    if not (sys.version_info[0] >= 3 and sys.version_info[1] >= 6):
        print("""
     MindsDB server requires Python >= 3.7 to run

     Once you have Python 3.7 installed you can tun mindsdb as follows:

     1. create and activate venv:
     python3.7 -m venv venv
     source venv/bin/activate

     2. install MindsDB:
     pip3 install mindsdb

     3. Run MindsDB
     python3.7 -m mindsdb

     More instructions in https://docs.mindsdb.com
         """)
        exit(1)

    # --- VERSION MODE ----
    if args is not None and args.version:
        print(f'MindsDB {mindsdb_version}')
        sys.exit(0)

    # --- MODULE OR LIBRARY IMPORT MODE ----
    if args is not None and args.config is not None:
        config_path = args.config
        with open(config_path, 'r') as fp:
            user_config = json.load(fp)
    else:
        user_config = {}
        config_path = 'absent'
    os.environ['MINDSDB_CONFIG_PATH'] = config_path

    mindsdb_config = Config()
    create_dirs_recursive(mindsdb_config['paths'])

    if telemetry_file_exists(mindsdb_config['storage_dir']):
        os.environ['CHECK_FOR_UPDATES'] = '0'
        print('\n x telemetry disabled! \n')
    elif os.getenv('CHECK_FOR_UPDATES', '1').lower() in ['0', 'false', 'False'] or mindsdb_config.get('cloud', False):
        disable_telemetry(mindsdb_config['storage_dir'])
        print('\n x telemetry disabled \n')
    else:
        print('\n ✓ telemetry enabled \n')

    if os.environ.get('FLASK_SECRET_KEY') is None:
        os.environ['FLASK_SECRET_KEY'] = secrets.token_hex(32)

    # -------------------------------------------------------

    # initialization
    db.init()
    log.initialize_log()

    mp.freeze_support()
    config = Config()

    environment = config.get('environment')
    if environment == 'aws_marketplace':
        try:
            register_oauth_client()
        except Exception as e:
            print(f'Something went wrong during client register: {e}')
    elif environment != 'local':
        try:
            aws_meta_data = get_aws_meta_data()
            Config().update({
                'aws_meta_data': aws_meta_data
            })
        except Exception:
            pass

    is_cloud = config.get('cloud', False)
    # need configure migration behavior by env_variables
    # leave 'is_cloud' for now, but needs to be removed further
    run_migration_separately = os.environ.get("SEPARATE_MIGRATIONS", False)
    if run_migration_separately in (False, "false", "False", 0, "0", ""):
        run_migration_separately = False
    else:
        run_migration_separately = True

    if not is_cloud and not run_migration_separately:
        print('Applying database migrations:')
        try:
            from mindsdb.migrations import migrate
            migrate.migrate_to_head()
        except Exception as e:
            print(f'Error! Something went wrong during DB migrations: {e}')

    if args.verbose is True:
        # Figure this one out later
        pass

    if args.install_handlers is not None:
        handlers_list = [s.strip() for s in args.install_handlers.split(',')]
        # import_meta = handler_meta.get('import', {})
        for handler_name, handler_meta in integration_controller.get_handlers_import_status().items():
            if handler_name not in handlers_list:
                continue
            import_meta = handler_meta.get('import', {})
            if import_meta.get('success') is True:
                print(f"{'{0: <18}'.format(handler_name)} - already installed")
                continue
            result = install_dependencies(import_meta.get('dependencies', []))
            if result.get('success') is True:
                print(f"{'{0: <18}'.format(handler_name)} - successfully installed")
            else:
                print(f"{'{0: <18}'.format(handler_name)} - error during dependencies installation: {result.get('error_message', 'unknown error')}")
        sys.exit(0)

    print(f'Version {mindsdb_version}')
    print(f'Configuration file:\n   {config.config_path}')
    print(f"Storage path:\n   {config['paths']['root']}")

    for handler_name, handler_meta in integration_controller.get_handlers_import_status().items():
        import_meta = handler_meta.get('import', {})
        if import_meta.get('success', False) is not True:
            print(dedent('''
                Some handlers cannot be imported. You can check list of available handlers by execute command in sql editor:
                    select * from information_schema.handlers;
            '''))
            break

    if not is_cloud:
        # region creating permanent integrations
        for integration_name, handler in integration_controller.get_handlers_import_status().items():
            if handler.get('permanent'):
                integration_meta = integration_controller.get(name=integration_name)
                if integration_meta is None:
                    integration_record = db.Integration(
                        name=integration_name,
                        data={},
                        engine=integration_name,
                        company_id=None
                    )
                    db.session.add(integration_record)
                    db.session.commit()
        # endregion

        # region Mark old predictors as outdated
        is_modified = False
        predictor_records = db.session.query(db.Predictor).filter(db.Predictor.deleted_at.is_(None)).all()
        if len(predictor_records) > 0:
            sucess, compatible_versions = get_versions_where_predictors_become_obsolete()
            if sucess is True:
                compatible_versions = [version.parse(x) for x in compatible_versions]
                mindsdb_version_parsed = version.parse(mindsdb_version)
                compatible_versions = [x for x in compatible_versions if x <= mindsdb_version_parsed]
                if len(compatible_versions) > 0:
                    last_compatible_version = compatible_versions[-1]
                    for predictor_record in predictor_records:
                        if (
                            isinstance(predictor_record.mindsdb_version, str)
                            and version.parse(predictor_record.mindsdb_version) < last_compatible_version
                        ):
                            predictor_record.update_status = 'available'
                            is_modified = True
        if is_modified is True:
            db.session.commit()
        # endregion

    if args.api is None:
        api_arr = ['http', 'mysql']
    else:
        api_arr = args.api.split(',')

    with_nlp = False
    if 'nlp' in api_arr:
        with_nlp = True
        api_arr.remove('nlp')

    apis = {
        api: {
            'port': config['api'][api]['port'],
            'process': None,
            'started': False
        } for api in api_arr
    }

    start_functions = {
        'http': start_http,
        'mysql': start_mysql,
        'mongodb': start_mongo,
        'postgres': start_postgres,
        'jobs': start_scheduler,
        'chatbot': start_chatbot
    }

    if config.get('jobs', {}).get('disable') is not True:
        apis['jobs'] = {
            'process': None,
            'started': False
        }

    # disabled on cloud
    if config.get('chatbot', {}).get('disable') is not True and not is_cloud:
        apis['chatbot'] = {
            'process': None,
            'started': False
        }

    ctx = mp.get_context('spawn')
    for api_name, api_data in apis.items():
        if api_data['started']:
            continue
        print(f'{api_name} API: starting...')
        try:
            if api_name == 'http':
                p = ctx.Process(target=start_functions[api_name], args=(args.verbose, args.no_studio, with_nlp))
            else:
                p = ctx.Process(target=start_functions[api_name], args=(args.verbose,))
            p.start()
            api_data['process'] = p
        except Exception as e:
            log.logger.error(f'Failed to start {api_name} API with exception {e}\n{traceback.format_exc()}')
            close_api_gracefully(apis)
            raise e

    atexit.register(close_api_gracefully, apis=apis)

    async def wait_api_start(api_name, pid, port):
        timeout = 60
        start_time = time.time()
        started = is_pid_listen_port(pid, port)
        while (time.time() - start_time) < timeout and started is False:
            await asyncio.sleep(0.5)
            started = is_pid_listen_port(pid, port)
        return api_name, port, started

    async def wait_apis_start():
        futures = [
            wait_api_start(api_name, api_data['process'].pid, api_data['port'])
            for api_name, api_data in apis.items() if 'port' in api_data
        ]
        for i, future in enumerate(asyncio.as_completed(futures)):
            api_name, port, started = await future
            if started:
                print(f"{api_name} API: started on {port}")
            else:
                log.logger.error(f"ERROR: {api_name} API cant start on {port}")

    async def join_process(process, name):
        try:
            process.join()
        except KeyboardInterrupt:
            print('Got keyboard interrupt, stopping APIs')
            close_api_gracefully(apis)
        finally:
            print(f'{name} API: stopped')

    async def gather_apis():
        await asyncio.gather(
            *[join_process(api_data['process'], api_name) for api_name, api_data in apis.items()],
            return_exceptions=False
        )

    ioloop = asyncio.new_event_loop()
    ioloop.run_until_complete(wait_apis_start())

    threading.Thread(target=do_clean_process_marks).start()

    ioloop.run_until_complete(gather_apis())
    ioloop.close()



from mindsdb.api.mongo.classes import Responder
import mindsdb.api.mongo.functions as helpers


class Responce(Responder):
    when = {'dbStats': helpers.is_true}

    def result(self, query, request_env, mindsdb_env, session):
        db = query['$db']
        collections = 0
        if db == 'mindsdb':
            collections = 2 + len(mindsdb_env['model_controller'].get_models())
        return {
            'db': db,
            'collections': collections,
            'views': 0,
            'objects': 0,
            'avgObjSize': 0,
            'dataSize': 0,
            'storageSize': 0,
            'numExtents': 0,
            'indexes': 0,
            'indexSize': 0,
            'fileSize': 0,
            'fsUsedSize': 0,
            'fsTotalSize': 0,
            'ok': 1
        }


responder = Responce()



from mindsdb.integrations.libs.const import HANDLER_TYPE

try:
    from .snowflake_handler import SnowflakeHandler as Handler
    import_error = None
except Exception as e:
    Handler = None
    import_error = e
from .__about__ import __version__ as version, __description__ as description


title = 'Snowflake'
name = 'snowflake'
type = HANDLER_TYPE.DATA

__all__ = [
    'Handler', 'version', 'name', 'type', 'title', 'description', 'icon_path'
]



import os
from pathlib import Path

TELEMETRY_FILE = 'telemetry.lock'


def enable_telemetry(storage_dir):
    os.environ['CHECK_FOR_UPDATES'] = '1'
    path = os.path.join(storage_dir, TELEMETRY_FILE)
    if os.path.exists(path):
        os.remove(path)


def disable_telemetry(storage_dir):
    os.environ['CHECK_FOR_UPDATES'] = '0'
    path = os.path.join(storage_dir, TELEMETRY_FILE)
    with open(path, 'w') as _:
        pass


def telemetry_file_exists(storage_dir):
    path = os.path.join(storage_dir, TELEMETRY_FILE)
    return os.path.exists(path)


def inject_telemetry_to_static(static_folder):
    TEXT = '<script>localStorage.isTestUser = true;</script>'
    index = Path(static_folder).joinpath('index.html')
    disable_telemetry = os.getenv('CHECK_FOR_UPDATES', '1').lower() in ['0', 'false', 'False']
    if index.is_file():
        with open(str(index), 'rt') as f:
            content = f.read()
        script_index = content.find('<script>')
        need_update = True
        if TEXT not in content and disable_telemetry:
            content = content[:script_index] + TEXT + content[script_index:]
        elif not disable_telemetry and TEXT in content:
            content = content.replace(TEXT, '')
        else:
            need_update = False

        if need_update:
            with open(str(index), 'wt') as f:
                f.write(content)



import os
from typing import Optional, Dict

from anthropic import Anthropic, HUMAN_PROMPT, AI_PROMPT
import pandas as pd

from mindsdb.utilities.config import Config
from mindsdb.integrations.libs.base import BaseMLEngine

from mindsdb.utilities.log import get_log


logger = get_log("integrations.anthropic_handler")

class AnthropicHandler(BaseMLEngine):
    """
    Integration with the Anthropic LLM Python Library
    """
    name = 'anthropic'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.default_chat_model = 'claude-2'
        self.supported_chat_models = ['claude-1', 'claude-2']
        self.default_max_tokens = 100
        self.generative = True
        self.connection = None

    def create(self, target: str, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:


        if 'using' not in args:
            raise Exception("Anthropic engine requires a USING clause! Refer to its documentation for more details.")

        if 'model' not in args['using']:
            args['using']['model'] = self.default_chat_model
        elif args['using']['model'] not in self.supported_chat_models:
            raise Exception(f"Invalid chat model. Please use one of {self.supported_chat_models}")

        if 'max_tokens' not in args['using']:
            args['using']['max_tokens'] = self.default_max_tokens
        
        self.model_storage.json_set('args', args)

    def predict(self, df: Optional[pd.DataFrame] = None, args: Optional[Dict] = None) -> None:

        args = self.model_storage.json_get('args')
        api_key = self._get_anthropic_api_key(args)

        self.connection = Anthropic(api_key=api_key,)

        input_column = args['using']['column']

        if input_column not in df.columns:
            raise RuntimeError(f'Column "{input_column}" not found in input data')
        
        result_df = pd.DataFrame() 

        result_df['predictions'] = df[input_column].apply(self.predict_answer)     

        result_df = result_df.rename(columns={'predictions': args['target']})
        
        return result_df


    def _get_anthropic_api_key(self, args, strict=True):
        """ 
        API_KEY preference order:
            1. provided at model creation
            2. provided at engine creation
            3. ANTHROPIC_API_KEY env variable
            4. anthropic.api_key setting in config.json
        """

        # 1
        if 'api_key' in args['using']:
            return args['using']['api_key']
        # 2
        connection_args = self.engine_storage.get_connection_args()
        if 'api_key' in connection_args:
            return connection_args['api_key']
        # 3
        api_key = os.getenv('ANTHROPIC_API_KEY')
        if api_key is not None:
            return api_key
        # 4
        config = Config()
        anthropic_cfg = config.get('anthropic', {})
        if 'api_key' in anthropic_cfg:
            return anthropic_cfg['api_key']

        if strict:
            raise Exception(f'Missing API key "api_key". Either re-create this ML_ENGINE specifying the `api_key` parameter,\
                 or re-create this model and pass the API key with `USING` syntax.')  

    def predict_answer(self,text):
        """ 
        connects with anthropic api to predict the answer for the particular question

        """ 

        args = self.model_storage.json_get('args')

        completion = self.connection.completions.create(
            model=args['using']['model'],
            max_tokens_to_sample=args['using']['max_tokens'],
            prompt=f"{HUMAN_PROMPT} {text} {AI_PROMPT}",
        )

        return completion.completion
